""" Shared fixtures for tests """

import logging
from pathlib import Path
from shutil import copytree
import os

import numpy as np
import pandas as pd
import pytest

import resqpy.grid as grr
import resqpy.olio.xml_et as rqet
import resqpy.property as rqp
from resqpy.crs import Crs
from resqpy.model import Model, new_model
from resqpy.organize import WellboreFeature, WellboreInterpretation
from resqpy.well import MdDatum, Trajectory, WellboreFrame
import resqpy.time_series as rqts


@pytest.fixture(autouse = True)
def capture_logs(caplog):
    """Always capture log messages from respy"""

    caplog.set_level(logging.DEBUG, logger = "resqpy")


@pytest.fixture
def tmp_model(tmp_path):
    """Example resqpy model in a temporary directory unique to each test"""

    return new_model(str(tmp_path / 'tmp_model.epc'))


@pytest.fixture
def example_model_and_crs(tmp_model):
    """ Returns a fresh RESQML Model and Crs, in a temporary directory """

    # TODO: parameterise with m or feet
    xyz_uom = 'm'

    # Create a model with a coordinate reference system
    crs = Crs(parent_model = tmp_model, z_inc_down = True, xy_units = xyz_uom, z_units = xyz_uom)
    crs.create_xml()

    return tmp_model, crs


@pytest.fixture
def example_model_with_well(example_model_and_crs):
    """ Model with a single well with a vertical trajectory """

    wellname = 'well A'
    elevation = 100
    md_uom = 'm'

    model, crs = example_model_and_crs

    # Create a single well feature and interpretation
    well_feature = WellboreFeature(parent_model = model, feature_name = wellname)
    well_feature.create_xml()
    well_interp = WellboreInterpretation(parent_model = model, wellbore_feature = well_feature, is_drilled = True)
    well_interp.create_xml()

    # Create a measured depth datum
    datum = MdDatum(parent_model = model,
                    crs_uuid = crs.uuid,
                    location = (0, 0, -elevation),
                    md_reference = 'kelly bushing')
    datum.create_xml()

    # Create trajectory of a vertical well
    mds = np.array([0, 1000, 2000])
    zs = mds - elevation
    traj = Trajectory(parent_model = model,
                      md_datum = datum,
                      data_frame = pd.DataFrame(dict(MD = mds, X = 0, Y = 0, Z = zs)),
                      length_uom = md_uom,
                      represented_interp = well_interp)
    traj.write_hdf5(mode = 'w')
    traj.create_xml()

    return model, well_interp, datum, traj


@pytest.fixture
def example_model_with_logs(example_model_with_well):

    model, well_interp, datum, traj = example_model_with_well

    frame = WellboreFrame(
        model,
        trajectory = traj,
        represented_interp = well_interp,
        mds = [1, 2, 3, 4],
    )
    frame.write_hdf5()
    frame.create_xml(title = 'Log run A')

    log_collection = frame.logs
    log_collection.add_log("GR", [1, 2, 1, 2], 'gAPI')
    log_collection.add_log("NPHI", [0.1, 0.1, np.NaN, np.NaN], 'v/v')

    return model, well_interp, datum, traj, frame, log_collection


@pytest.fixture
def test_data_path(tmp_path):
    """ Return pathlib.Path pointing to temporary copy of tests/example_data

   Use a fresh temporary directory for each test.
   """
    master_path = (Path(__file__) / '../test_data').resolve()
    data_path = Path(tmp_path) / 'test_data'

    assert master_path.exists()
    assert not data_path.exists()
    copytree(str(master_path), str(data_path))
    return data_path


@pytest.fixture
def example_model_with_properties(tmp_path):
    """Model with a grid (5x5x3) and properties.
   Properties:
   - Zone (discrete)
   - VPC (discrete)
   - Fault block (discrete)
   - Facies (discrete)
   - NTG (continuous)
   - POR (continuous)
   - SW (continuous)
   """
    model_path = str(tmp_path / 'test_no_rels.epc')
    model = Model(create_basics = True, create_hdf5_ext = True, epc_file = model_path, new_epc = True)
    model.store_epc(model.epc_file)

    grid = grr.RegularGrid(parent_model = model,
                           origin = (0, 0, 0),
                           extent_kji = (3, 5, 5),
                           crs_uuid = rqet.uuid_for_part_root(model.crs_root),
                           set_points_cached = True)
    grid.cache_all_geometry_arrays()
    grid.write_hdf5_from_caches(file = model.h5_file_name(file_must_exist = False), mode = 'w')

    grid.create_xml(ext_uuid = model.h5_uuid(),
                    title = 'grid',
                    write_geometry = True,
                    add_cell_length_properties = False)
    model.store_epc()

    zone = np.ones(shape = (5, 5))
    zone_array = np.array([zone, zone + 1, zone + 2], dtype = 'int')

    vpc = np.array([[1, 1, 1, 2, 2], [1, 1, 1, 2, 2], [1, 1, 1, 2, 2], [1, 1, 1, 2, 2], [1, 1, 1, 2, 2]])
    vpc_array = np.array([vpc, vpc, vpc])

    facies = np.array([[1, 1, 1, 2, 2], [1, 1, 2, 2, 2], [1, 2, 2, 2, 3], [2, 2, 2, 3, 3], [2, 2, 3, 3, 3]])
    facies_array = np.array([facies, facies, facies])

    fb = np.array([[1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [2, 2, 2, 2, 2], [2, 2, 2, 2, 2]])
    fb_array = np.array([fb, fb, fb])

    ntg = np.array([[0, 0.5, 0, 0.5, 0], [0.5, 0, 0.5, 0, 0.5], [0, 0.5, 0, 0.5, 0], [0.5, 0, 0.5, 0, 0.5],
                    [0, 0.5, 0, 0.5, 0]])
    ntg_array = np.array([ntg, ntg, ntg])

    por = np.array([[1, 1, 1, 1, 1], [0.5, 0.5, 0.5, 0.5, 0.5], [1, 1, 1, 1, 1], [0.5, 0.5, 0.5, 0.5, 0.5],
                    [1, 1, 1, 1, 1]])
    por_array = np.array([por, por, por])

    sat = np.array([[1, 0.5, 1, 0.5, 1], [1, 0.5, 1, 0.5, 1], [1, 0.5, 1, 0.5, 1], [1, 0.5, 1, 0.5, 1],
                    [1, 0.5, 1, 0.5, 1]])
    sat_array = np.array([sat, sat, sat])

    perm = np.array([[1, 10, 10, 100, 100], [1, 10, 10, 100, 100], [1, 10, 10, 100, 100], [1, 10, 10, 100, 100],
                     [1, 10, 10, 100, 100]])
    perm_array = np.array([perm, perm, perm], dtype = 'float')

    collection = rqp.GridPropertyCollection()
    collection.set_grid(grid)
    for array, name, kind, discrete, facet_type, facet in zip(
        [zone_array, vpc_array, fb_array, facies_array, ntg_array, por_array, sat_array, perm_array],
        ['Zone', 'VPC', 'Fault block', 'Facies', 'NTG', 'POR', 'SW', 'Perm'], [
            'discrete', 'discrete', 'discrete', 'discrete', 'net to gross ratio', 'porosity', 'saturation',
            'permeability rock'
        ], [True, True, True, True, False, False, False, False],
        [None, None, None, None, None, None, None, 'direction'], [None, None, None, None, None, None, None, 'I']):
        collection.add_cached_array_to_imported_list(cached_array = array,
                                                     source_info = '',
                                                     keyword = name,
                                                     discrete = discrete,
                                                     uom = None,
                                                     time_index = None,
                                                     null_value = None,
                                                     property_kind = kind,
                                                     facet_type = facet_type,
                                                     facet = facet,
                                                     realization = None)
        collection.write_hdf5_for_imported_list()
        collection.create_xml_for_imported_list_and_add_parts_to_model()
    model.store_epc()

    return model


@pytest.fixture
def example_model_with_prop_ts_rels(tmp_path):
    """Model with a grid (5x5x3) and properties.
   Properties:
   - Zone (discrete)
   - VPC (discrete)
   - Fault block (discrete)
   - Facies (discrete)
   - NTG (continuous)
   - POR (continuous)
   - SW (continuous) (recurrent)
   """
    model_path = str(tmp_path / 'test_model.epc')
    model = Model(create_basics = True, create_hdf5_ext = True, epc_file = model_path, new_epc = True)
    model.store_epc(model.epc_file)

    grid = grr.RegularGrid(parent_model = model,
                           origin = (0, 0, 0),
                           extent_kji = (3, 5, 5),
                           crs_uuid = rqet.uuid_for_part_root(model.crs_root),
                           set_points_cached = True)
    grid.cache_all_geometry_arrays()
    grid.write_hdf5_from_caches(file = model.h5_file_name(file_must_exist = False), mode = 'w')

    grid.create_xml(ext_uuid = model.h5_uuid(),
                    title = 'grid',
                    write_geometry = True,
                    add_cell_length_properties = False)
    model.store_epc()

    zone = np.ones(shape = (5, 5), dtype = 'int')
    zone_array = np.array([zone, zone + 1, zone + 2], dtype = 'int')

    vpc = np.array([[1, 1, 1, 2, 2], [1, 1, 1, 2, 2], [1, 1, 1, 2, 2], [1, 1, 1, 2, 2], [1, 1, 1, 2, 2]], dtype = 'int')
    vpc_array = np.array([vpc, vpc, vpc], dtype = 'int')

    facies = np.array([[1, 1, 1, 2, 2], [1, 1, 2, 2, 2], [1, 2, 2, 2, 3], [2, 2, 2, 3, 3], [2, 2, 3, 3, 3]],
                      dtype = 'int')
    facies_array = np.array([facies, facies, facies], dtype = 'int')

    perm = np.array([[1, 1, 1, 10, 10], [1, 1, 1, 10, 10], [1, 1, 1, 10, 10], [1, 1, 1, 10, 10], [1, 1, 1, 10, 10]])
    perm_array = np.array([perm, perm, perm], dtype = 'float')

    fb = np.array([[1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [2, 2, 2, 2, 2], [2, 2, 2, 2, 2]], dtype = 'int')
    fb_array = np.array([fb, fb, fb], dtype = 'int')

    ntg = np.array([[0, 0.5, 0, 0.5, 0], [0.5, 0, 0.5, 0, 0.5], [0, 0.5, 0, 0.5, 0], [0.5, 0, 0.5, 0, 0.5],
                    [0, 0.5, 0, 0.5, 0]])
    ntg1_array = np.array([ntg, ntg, ntg])
    ntg2_array = np.array([ntg + 0.1, ntg + 0.1, ntg + 0.1])

    por = np.array([[1, 1, 1, 1, 1], [0.5, 0.5, 0.5, 0.5, 0.5], [1, 1, 1, 1, 1], [0.5, 0.5, 0.5, 0.5, 0.5],
                    [1, 1, 1, 1, 1]])
    por1_array = np.array([por, por, por])
    por2_array = np.array([por - 0.1, por - 0.1, por - 0.1])

    sat = np.array([[1, 0.5, 1, 0.5, 1], [1, 0.5, 1, 0.5, 1], [1, 0.5, 1, 0.5, 1], [1, 0.5, 1, 0.5, 1],
                    [1, 0.5, 1, 0.5, 1]])
    sat1_array = np.array([sat, sat, sat])
    sat2_array = np.array([sat, sat, np.where(sat == 0.5, 0.75, sat)])
    sat3_array = np.array(
        [np.where(sat == 0.5, 0.75, sat),
         np.where(sat == 0.5, 0.75, sat),
         np.where(sat == 0.5, 0.75, sat)])

    collection = rqp.GridPropertyCollection()
    collection.set_grid(grid)

    ts = rqts.TimeSeries(parent_model = model, first_timestamp = '2000-01-01Z')
    ts.extend_by_days(365)
    ts.extend_by_days(365)

    ts.create_xml()

    lookup = rqp.StringLookup(parent_model = model, int_to_str_dict = {1: 'channel', 2: 'interbedded', 3: 'shale'})
    lookup.create_xml()

    model.store_epc()

    # Add non-varying properties
    for array, name, kind, discrete, facet_type, facet in zip([zone_array, vpc_array, fb_array, perm_array],
                                                              ['Zone', 'VPC', 'Fault block', 'Perm'],
                                                              ['discrete', 'discrete', 'discrete', 'permeability rock'],
                                                              [True, True, True, False],
                                                              [None, None, None, 'direction'], [None, None, None, 'J']):
        collection.add_cached_array_to_imported_list(cached_array = array,
                                                     source_info = '',
                                                     keyword = name,
                                                     discrete = discrete,
                                                     uom = None,
                                                     time_index = None,
                                                     null_value = None,
                                                     property_kind = kind,
                                                     facet_type = facet_type,
                                                     facet = facet,
                                                     realization = None)
        collection.write_hdf5_for_imported_list()
        collection.create_xml_for_imported_list_and_add_parts_to_model()

    # Add realisation varying properties
    for array, name, kind, rel in zip([ntg1_array, por1_array, ntg2_array, por2_array], ['NTG', 'POR', 'NTG', 'POR'],
                                      ['net to gross ratio', 'porosity', 'net to gross ratio', 'porosity'],
                                      [0, 0, 1, 1]):
        collection.add_cached_array_to_imported_list(cached_array = array,
                                                     source_info = '',
                                                     keyword = name,
                                                     discrete = False,
                                                     uom = None,
                                                     time_index = None,
                                                     null_value = None,
                                                     property_kind = kind,
                                                     facet_type = None,
                                                     facet = None,
                                                     realization = rel)
        collection.write_hdf5_for_imported_list()
        collection.create_xml_for_imported_list_and_add_parts_to_model()

    # Add categorial property
    collection.add_cached_array_to_imported_list(cached_array = facies_array,
                                                 source_info = '',
                                                 keyword = 'Facies',
                                                 discrete = True,
                                                 uom = None,
                                                 time_index = None,
                                                 null_value = None,
                                                 property_kind = 'discrete',
                                                 facet_type = None,
                                                 facet = None,
                                                 realization = None)
    collection.write_hdf5_for_imported_list()
    collection.create_xml_for_imported_list_and_add_parts_to_model(string_lookup_uuid = lookup.uuid)

    # Add time varying properties
    for array, ts_index in zip([sat1_array, sat2_array, sat3_array], [0, 1, 2]):
        collection.add_cached_array_to_imported_list(cached_array = array,
                                                     source_info = '',
                                                     keyword = 'SW',
                                                     discrete = False,
                                                     uom = None,
                                                     time_index = ts_index,
                                                     null_value = None,
                                                     property_kind = 'saturation',
                                                     facet_type = 'what',
                                                     facet = 'water',
                                                     realization = None)
        collection.write_hdf5_for_imported_list()
        collection.create_xml_for_imported_list_and_add_parts_to_model(time_series_uuid = ts.uuid)
    model.store_epc()

    return model


@pytest.fixture
def example_model_and_cellio(example_model_and_crs, tmp_path):
    model, crs = example_model_and_crs
    grid = grr.RegularGrid(model,
                           extent_kji = (3, 4, 3),
                           dxyz = (50.0, -50.0, 50.0),
                           origin = (0.0, 0.0, 100.0),
                           crs_uuid = crs.uuid,
                           set_points_cached = True)

    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    grid_uuid = grid.uuid
    cellio_file = os.path.join(model.epc_directory, 'cellio.dat')
    well_name = 'Banoffee'
    source_df = pd.DataFrame(
        [[2, 2, 1, 25, -25, 125, 26, -26, 126], [2, 2, 2, 26, -26, 126, 27, -27, 127],
         [2, 2, 3, 27, -27, 127, 28, -28, 128]],
        columns = ['i_index', 'j_index', 'k_index', 'x_in', 'y_in', 'z_in', 'x_out', 'y_out', 'z_out'])

    with open(cellio_file, 'w') as fp:
        fp.write('1.0\n')
        fp.write('Undefined\n')
        fp.write(f'{well_name}\n')
        fp.write('9\n')
        for col in source_df.columns:
            fp.write(f' {col}\n')
        for row_index in range(len(source_df)):
            row = source_df.iloc[row_index]
            for col in source_df.columns:
                fp.write(f' {int(row[col])}')
            fp.write('\n')

    return model, cellio_file, well_name



import datetime

from resqpy.time_series._from_nexus_summary import _process_summary_entries, time_series_from_nexus_summary


def test_process_summary_entries_none():
    # arrange
    summary_entries = []
    # act
    result = _process_summary_entries(summary_entries)
    # assert
    assert result is None


# (time step no., cumulative time (days), date (dd/mm/yyyy))


def test_process_summary_entries_first():
    # arrange
    summary_entries = [(0, 1, datetime.datetime.strptime('24052010', '%d%m%Y').date())]
    # act
    result = _process_summary_entries(summary_entries)
    # assert
    assert len(result.timestamps) == 1
    assert result.timestamps[0] == '2010-05-24T00:00:00Z'


def test_process_summary_entries_delta():
    # arrange
    summary_entries = [(1, 1, datetime.datetime.strptime('24052010', '%d%m%Y').date())]
    # act
    result = _process_summary_entries(summary_entries)
    # assert
    assert len(result.timestamps) == 2
    assert result.timestamps[1] == '2010-05-24T00:00:00Z'
    assert result.timestamps[0] == '2010-05-23T00:00:00Z'


def test_process_summary_entries_out_of_sequence():
    # arrange
    summary_entries = [(0, 1, datetime.datetime.strptime('24052010', '%d%m%Y').date()),
                       (0, 1, datetime.datetime.strptime('24052010', '%d%m%Y').date())]
    # act
    result = _process_summary_entries(summary_entries)
    # assert
    assert len(result.timestamps) == 1
    assert result.timestamps[0] == '2010-05-24T00:00:00Z'


def test_process_summary_entries_out_of_sequence_v2():
    # arrange
    summary_entries = [(1, 1, datetime.datetime.strptime('24052010', '%d%m%Y').date()),
                       (3, 4, datetime.datetime.strptime('28052010', '%d%m%Y').date()),
                       (2, 2, datetime.datetime.strptime('26052010', '%d%m%Y').date())]
    # act
    result = _process_summary_entries(summary_entries)
    # assert
    assert len(result.timestamps) == 4
    assert result.timestamps[0] == '2010-05-23T00:00:00Z'
    assert result.timestamps[1] == '2010-05-24T00:00:00Z'
    assert result.timestamps[2] == '2010-05-25T00:00:00Z'
    assert result.timestamps[3] == '2010-05-27T00:00:00Z'


def test_process_summary_entries_in_sequence():
    # arrange
    summary_entries = [(1, 1, datetime.datetime.strptime('24052010', '%d%m%Y').date()),
                       (3, 4, datetime.datetime.strptime('28052010', '%d%m%Y').date())]
    # act
    result = _process_summary_entries(summary_entries)
    # assert
    assert len(result.timestamps) == 4
    assert result.timestamps[0] == '2010-05-23T00:00:00Z'
    assert result.timestamps[1] == '2010-05-24T00:00:00Z'
    assert result.timestamps[2] == '2010-05-24T00:00:00Z'
    assert result.timestamps[3] == '2010-05-27T00:00:00Z'


def test_process_summary_entries_incorrect_days():
    # arrange
    summary_entries = [(1, 1, datetime.datetime.strptime('24052010', '%d%m%Y').date()),
                       (2, 0, datetime.datetime.strptime('28052010', '%d%m%Y').date())]
    # act
    result = _process_summary_entries(summary_entries)
    # assert
    assert len(result.timestamps) == 3
    assert result.timestamps[0] == '2010-05-23T00:00:00Z'
    assert result.timestamps[1] == '2010-05-24T00:00:00Z'
    assert result.timestamps[2] == '2010-05-24T00:00:00Z'


def test_time_series_from_nexus_summary_is_not_file(mocker):
    # arrange
    is_file_mock = mocker.MagicMock(return_value = False)
    mocker.patch('os.path.isfile', is_file_mock)
    # log_warning_mock = mocker.MagicMock()
    # logging_mock = mocker.MagicMock()
    # logging_mock.getlogger.warning = log_warning_mock
    # mocker.patch(logging, logging_mock)
    # act
    result = time_series_from_nexus_summary('test')
    # assert
    assert result is None
    # log_warning_mock.assert_called_once_with('Summary file not found: ' + 'test')


def test_time_series_from_nexus_summary_if_no_file():
    # arrange
    # act
    result = time_series_from_nexus_summary('')
    # assert
    assert result is None


# def test_time_series_from_nexus_summary_open_file(mocker):
#     # arrange
#     is_file_mock = mocker.MagicMock(return_value=True)
#     mocker.patch('os.path.isfile', is_file_mock)
#     summary_entries = [(1, 1, datetime.datetime.strptime('24052010', '%d%m%Y').date()),
#                        (2, 0, datetime.datetime.strptime('28052010', '%d%m%Y').date())]
#     open_file_mock = mocker.MagicMock(return_value=summary_entries)
#     mocker.patch(resqpy.time_series.time_series_from_nexus_summary.open_file, open_file_mock)
#     # act
#     result = time_series_from_nexus_summary('test')
#     # assert
#     assert result is None
#     open_file_mock.assert_called_once_with('test')



import pytest

import resqpy.well
from resqpy.model import Model


def test_MdDatum(example_model_and_crs):

    # Set up a new datum
    model, crs = example_model_and_crs
    epc = model.epc_file
    data = dict(
        location = (0, -99999, 3.14),
        md_reference = 'mean low water',
    )
    datum = resqpy.well.MdDatum(parent_model = model, crs_uuid = crs.uuid, **data)
    uuid = datum.uuid

    # Save to disk and reload
    datum.create_part()
    model.store_epc()

    del model, crs, datum
    model2 = Model(epc_file = epc)
    datum2 = resqpy.well.MdDatum(parent_model = model2, uuid = uuid)

    for key, expected_value in data.items():
        assert getattr(datum2, key) == expected_value, f"Issue with {key}"

    identical = resqpy.well.MdDatum(parent_model = model2, crs_uuid = datum2.crs_uuid, **data)
    data['md_reference'] = 'kelly bushing'
    different = resqpy.well.MdDatum(parent_model = model2, crs_uuid = datum2.crs_uuid, **data)
    assert identical == datum2
    assert different != datum2


@pytest.mark.parametrize('other_data,expected',
                         [(dict(location = (0, -99999, 3.14), md_reference = 'mean low water'), True),
                          (dict(location = (5, -30000, 5), md_reference = 'kelly bushing'), False)])
def test_is_equivalent(example_model_and_crs, other_data, expected):
    # Test that comparison of metadata items returns the expected result

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    data = dict(location = (0, -99999, 3.14), md_reference = 'mean low water')

    # --------- Act ----------
    datum1 = resqpy.well.MdDatum(parent_model = model, crs_uuid = crs.uuid, **data)
    datum2 = resqpy.well.MdDatum(parent_model = model, crs_uuid = crs.uuid, **other_data)
    result = datum1.is_equivalent(other = datum2)

    # --------- Assert ----------
    assert result is expected



import logging

import pytest

log = logging.getLogger(__name__)
log.debug('resqml_time_series.py version ')

from resqpy.time_series import TimeSeries


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_index_for_timestamp_not_later_than_later_timestamp(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.index_for_timestamp_not_later_than('2016-01-01T00:00:00.588Z')
    # assert
    assert result == 1


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_index_for_timestamp_not_later_than_same_timestamp(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.index_for_timestamp_not_later_than('2015-02-02T00:00:00.588Z')
    # assert
    assert result == 1


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_index_for_timestamp_not_later_is_none(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.index_for_timestamp_not_later_than('2013-02-02T00:00:00.588Z')
    # assert
    assert result is None


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_index_for_timestamp_not_earlier_than_earlier_timestamp(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.index_for_timestamp_not_earlier_than('2013-02-02T00:00:00.588Z')
    # assert
    assert result == 0


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_index_for_timestamp_not_earlier_than_later_timestamp(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.index_for_timestamp_not_earlier_than('2016-02-02T00:00:00.588Z')
    # assert
    assert result is None


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_index_for_timestamp_not_earlier_than_same_timestamp(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.index_for_timestamp_not_earlier_than('2015-02-02T00:00:00.588Z')
    # assert
    assert result == 1


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_index_for_timestamp_closest_to_empty_timestamps(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = None
    # act
    result = any_time_series.index_for_timestamp_closest_to('2000-01-01T00:00:00.588Z')
    # assert
    assert result is None


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_index_for_timestamp_closest_to_timestamp_too_early(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.index_for_timestamp_closest_to('2013-02-02T00:00:00.588Z')
    # assert
    assert result == 0


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_index_for_timestamp_closest_to_timestamp_len(mocker, timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    index_for_timestamp_not_later_mock = mocker.Mock(return_value = 1)
    any_time_series.index_for_timestamp_not_later_than = index_for_timestamp_not_later_mock
    # act
    result = any_time_series.index_for_timestamp_closest_to('2013-02-02T00:00:00.588Z')
    # assert
    assert result == 1


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_index_for_timestamp_closest_to_timestamp_equals_before(mocker, timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    index_for_timestamp_not_later_mock = mocker.Mock(return_value = 0)
    any_time_series.index_for_timestamp_not_later_than = index_for_timestamp_not_later_mock
    # act
    result = any_time_series.index_for_timestamp_closest_to('2014-01-01T00:00:00.588Z')
    # assert
    assert result == 0
    index_for_timestamp_not_later_mock.assert_called_once_with('2014-01-01T00:00:00.588Z')


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_index_for_timestamp_closest_to_timestamp_early_delta(mocker, timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    index_for_timestamp_not_later_mock = mocker.Mock(return_value = 0)
    any_time_series.index_for_timestamp_not_later_than = index_for_timestamp_not_later_mock
    # act
    result = any_time_series.index_for_timestamp_closest_to('2014-02-02T00:00:00.588Z')
    # assert
    assert result == 0
    index_for_timestamp_not_later_mock.assert_called_once_with('2014-02-02T00:00:00.588Z')


@pytest.mark.parametrize('timestamps',
                         [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z', '2016-03-02T00:00:00.588Z']])
def test_index_for_timestamp_closest_to_timestamp_later_delta(mocker, timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    index_for_timestamp_not_later_mock = mocker.Mock(return_value = 1)
    any_time_series.index_for_timestamp_not_later_than = index_for_timestamp_not_later_mock
    # act
    result = any_time_series.index_for_timestamp_closest_to('2016-01-01T00:00:00.588Z')
    # assert
    assert result == 2
    index_for_timestamp_not_later_mock.assert_called_once_with('2016-01-01T00:00:00.588Z')


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_duration_since_start_invalid_index(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.duration_since_start(-1)
    # assert
    assert result is None


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_duration_since_start_later_index(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.duration_since_start(2)
    # assert
    assert result is None


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_step_duration_invalid_index(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.step_duration(0)
    # assert
    assert result is None


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_step_duration_later_index(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.step_duration(2)
    # assert
    assert result is None


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2014-01-03T00:00:00.588Z']])
def test_step_duration_between_index(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.step_duration(1)
    # assert
    assert result.duration.days == 2


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2014-01-03T00:00:00.588Z']])
def test_step_days_none(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.step_days(0)
    # assert
    assert result is None


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2014-01-03T00:00:00.588Z']])
def test_step_days_number(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.step_days(1)
    # assert
    assert result == 2


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2014-01-03T00:00:00.588Z']])
def test_add_timestamp_before(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    new_timestamp = '2013-01-01T00:00:00.588Z'
    # act
    any_time_series.add_timestamp(new_timestamp, allow_insertion = True)
    # assert
    assert any_time_series.timestamps[0] == new_timestamp
    assert any_time_series.timestamps[1] == '2014-01-01T00:00:00.588Z'
    assert any_time_series.timestamps[2] == '2014-01-03T00:00:00.588Z'


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2014-01-03T00:00:00.588Z']])
def test_add_timestamp_after(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    new_timestamp = '2015-01-01T00:00:00.588Z'
    # act
    any_time_series.add_timestamp(new_timestamp, allow_insertion = True)
    # assert
    assert any_time_series.timestamps[0] == '2014-01-01T00:00:00.588Z'
    assert any_time_series.timestamps[1] == '2014-01-03T00:00:00.588Z'
    assert any_time_series.timestamps[2] == new_timestamp


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2014-01-03T00:00:00.588Z']])
def test_add_timestamp_between(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = TimeSeries(model)
    any_time_series.timestamps = timestamps
    new_timestamp = '2014-01-02T00:00:00.588Z'
    # act
    any_time_series.add_timestamp(new_timestamp, allow_insertion = True)
    # assert
    assert any_time_series.timestamps[0] == '2014-01-01T00:00:00.588Z'
    assert any_time_series.timestamps[1] == new_timestamp
    assert any_time_series.timestamps[2] == '2014-01-03T00:00:00.588Z'



"""base_surface.py: base_surface class based on resqml standard."""

version = '4th November 2021'

# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company
# GOCAD is also a trademark of Emerson

import logging

log = logging.getLogger(__name__)

import resqpy.organize as rqo
from resqpy.olio.base import BaseResqpy


class BaseSurface(BaseResqpy):
    """Base class to implement shared methods for other classes in this module."""

    def create_interpretation_and_feature(self,
                                          kind = 'horizon',
                                          name = None,
                                          interp_title_suffix = None,
                                          is_normal = True):
        """Creates xml and objects for a represented interpretaion and interpreted feature, if not already present."""

        assert kind in ['horizon', 'fault', 'fracture', 'geobody boundary']
        assert name or self.title, 'title missing'
        if not name:
            name = self.title

        if self.represented_interpretation_root is not None:
            log.debug(f'represented interpretation already exisrs for surface {self.title}')
            return
        if kind in ['horizon', 'geobody boundary']:
            feature = rqo.GeneticBoundaryFeature(self.model, kind = kind, feature_name = name)
            feature.create_xml()
            if kind == 'horizon':
                interp = rqo.HorizonInterpretation(self.model, genetic_boundary_feature = feature, domain = 'depth')
            else:
                interp = rqo.GeobodyBoundaryInterpretation(self.model,
                                                           genetic_boundary_feature = feature,
                                                           domain = 'depth')
        elif kind in ['fault', 'fracture']:
            feature = rqo.TectonicBoundaryFeature(self.model, kind = kind, feature_name = name)
            feature.create_xml()
            interp = rqo.FaultInterpretation(self.model,
                                             is_normal = is_normal,
                                             tectonic_boundary_feature = feature,
                                             domain = 'depth')  # might need more arguments
        else:
            log.critical('code failure')
        interp_root = interp.create_xml(title_suffix = interp_title_suffix)
        self.set_represented_interpretation_root(interp_root)



"""Functions for reading zmap and roxar format files."""

version = '13th May 2021'

# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company

import logging

log = logging.getLogger(__name__)

import numpy as np


def read_zmap_header(inputfile):
    """Reads header lines from a zmap format file.

    returns:
       header_lines_count, no_rows, no_cols, minx, maxx, miny, maxy, null_value
    """

    # read header, read until second '@', record header lines and content
    with open(inputfile, 'r') as infile:
        comments = []
        head = []
        a = 0  # count '@'s
        headers = 0  # count header+comment lines
        while a < 2:
            line = infile.readline()
            headers += 1
            if line[:1] == '@':
                a += 1
            if line[:1] == '!':
                comments.append(line.strip())
            elif a == 1 or (line[:1] == '@'):
                row = line.strip().strip('@').split(',')
                head.append(row)
            else:
                log.error("Header section does not seem to be defined by 2 @s")
                return None, None, None
        line = infile.readline()
        if line[0] == '+':
            headers = headers + 1  # add extra line for the + symbol..

    for c in comments:
        log.debug(c)

    # ok now process the header
#  nodes_per_line                          = int(head[0][2])
#  field_w                                 = head[1][0]
    null_value = head[1][1].strip()
    null_value2 = head[1][2].strip()
    #  dec_places                              = head[1][3]
    #  strt_c                                  = head[1][4]
    no_rows = int(head[2][0])
    no_cols = int(head[2][1])
    minx = np.float64(head[2][2])
    maxx = np.float64(head[2][3])
    miny = np.float64(head[2][4])
    maxy = np.float64(head[2][5])

    # decide on the null value
    if not null_value:
        null_value = null_value2
    log.debug("Read {} header lines, we have {} rows, {} cols, and data from {} to {} and {} to {}".format(
        headers, no_rows, no_cols, minx, maxx, miny, maxy))

    return headers, no_rows, no_cols, minx, maxx, miny, maxy, null_value


# note: the RMS text format was previously known as the Roxar format


def read_roxar_header(inputfile):
    """Reads header lines from a roxar format file.

    returns:
       header_lines_count, no_rows, no_cols, minx, maxx, miny, maxy, null_value
    """

    with open(inputfile, 'r') as fp:
        words = fp.readline().split()
        assert words[0] == '-996', 'RMS text format indicator -996 not found'
        no_rows = int(words[1])
        # dx = float(words[2])
        # dy = float(words[3])
        words = fp.readline().split()
        minx = float(words[0])
        maxx = float(words[1])
        miny = float(words[2])
        maxy = float(words[3])
        words = fp.readline().split()
        no_cols = int(words[0])
    headers = 4
    null_value = '9999900.0000'
    return headers, no_rows, no_cols, minx, maxx, miny, maxy, null_value


def read_mesh(inputfile, dtype = np.float64, format = None):
    """Reads a mesh (lattice) from a zmap or roxar format file.

    returns:
       x, y, f: each a numpy float array of shape (no_rows, no_cols)
    """

    if format == 'zmap':
        headers, no_rows, no_cols, minx, maxx, miny, maxy, null_value = read_zmap_header(inputfile)
    elif format in ['rms', 'roxar']:
        headers, no_rows, no_cols, minx, maxx, miny, maxy, null_value = read_roxar_header(inputfile)
    else:
        raise ValueError('format not recognised for read_mesh: ' + str(format))
    # load the values in, converting null value to NaN's

    infile = open(inputfile, 'r')
    indata = infile.readlines()
    infile.close()
    c = 0
    f = np.zeros(no_cols * no_rows, dtype = dtype)
    n = 0
    for line in indata:
        c += 1
        if c <= headers:
            continue
        else:
            for x in line.split():
                if (x == null_value) or (dtype(x) == dtype(null_value)):
                    f[n] = np.nan
                else:
                    f[n] = dtype(x)
                n += 1
        if (n % 10000 == 0) and (n > 10000):
            log.debug("Read node {} of {}".format(n, no_cols * no_rows))

    log.debug("Read {} nodes".format(len(f)))

    if format == 'zmap':
        # reshape it, and swap axis. Beacuse columns major order from fortran.
        f = f.reshape((no_cols, no_rows)).swapaxes(0, 1)
    else:  # format in ['rms', 'roxar']
        f = f.reshape((no_rows, no_cols))

    # now generate x and y coords
    x = np.linspace(minx, maxx, no_cols)  # get x axis coords
    if format == 'zmap':
        y = np.linspace(maxy, miny, no_rows)
    else:  # format in ['rms', 'roxar']
        y = np.linspace(miny, maxy, no_rows)
    x, y = np.meshgrid(x, y)  # get x and y of every node
    assert x.shape == y.shape == f.shape

    return x, y, f


def read_zmapplusgrid(inputfile, dtype = np.float64):
    """Read zmapplus grid (surface mesh); returns triple (x, y, z) 2D arrays."""

    return read_mesh(inputfile, dtype = dtype, format = 'zmap')


def read_roxar_mesh(inputfile, dtype = np.float64):
    """Read RMS text format surface mesh; returns triple (x, y, z) 2D arrays.

    note:
       the RMS text format was previously known as the Roxar format
    """

    return read_mesh(inputfile, dtype = dtype, format = 'rms')


def read_rms_text_mesh(inputfile, dtype = np.float64):
    """Read RMS text format surface mesh; returns triple (x, y, z) 2D arrays.

    note:
       the RMS text format was previously known as the Roxar format
    """

    return read_roxar_mesh(inputfile, dtype = dtype)



"""Classes for RESQML objects related to surfaces."""

__all__ = ['BaseSurface', 'CombinedSurface', 'Mesh', 'TriangulatedPatch', 'PointSet', 'Surface']

from .base_surface import BaseSurface
from .combined_surface import CombinedSurface
from .mesh import Mesh
from .triangulated_patch import TriangulatedPatch
from .pointset import PointSet
from .surface import Surface

# Set "module" attribute of all public objects to this path.
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""_grids.py: functions supporting Model methods relating to grid objects."""

import resqpy.grid as grr
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet


def _root_for_ijk_grid(model, uuid = None, title = None):
    """Return xml root for IJK Grid part."""

    if title is not None:
        title = title.strip().upper()
    if uuid is None and not title:
        grid_root = model.root(obj_type = 'IjkGridRepresentation', title = 'ROOT', multiple_handling = 'oldest')
        if grid_root is None:
            grid_root = model.root(obj_type = 'IjkGridRepresentation')
    else:
        grid_root = model.root(obj_type = 'IjkGridRepresentation', uuid = uuid, title = title)

    assert grid_root is not None, 'IJK Grid part not found'

    return grid_root


def _resolve_grid_root(model, grid_root = None, uuid = None):
    """If grid root argument is None, returns the xml root for the IJK Grid part instead."""

    if grid_root is not None:
        if model.grid_root is None:
            model.grid_root = grid_root
    else:
        if model.grid_root is None:
            model.grid_root = _root_for_ijk_grid(model, uuid = uuid)
        grid_root = model.grid_root
    return grid_root


def _grid(model, title = None, uuid = None, find_properties = True):
    """Returns a shared Grid (or RegularGrid) object for this model, by default the 'main' grid."""

    if uuid is None and (title is None or title.upper() == 'ROOT'):
        if model.main_grid is not None:
            if find_properties:
                model.main_grid.extract_property_collection()
            return model.main_grid
        if title is None:
            grid_root = _resolve_grid_root(model)
        else:
            grid_root = _resolve_grid_root(model,
                                           grid_root = model.root(obj_type = 'IjkGridRepresentation', title = title))
    else:
        grid_root = model.root(obj_type = 'IjkGridRepresentation', uuid = uuid, title = title)
    assert grid_root is not None, 'IJK Grid part not found'
    if uuid is None:
        uuid = rqet.uuid_for_part_root(grid_root)
    for grid in model.grid_list:
        if grid.root is grid_root:
            if find_properties:
                grid.extract_property_collection()
            return grid
    grid = grr.any_grid(model, uuid = uuid, find_properties = find_properties)
    assert grid is not None, 'failed to instantiate grid object'
    if find_properties:
        grid.extract_property_collection()
    _add_grid(model, grid)
    return grid


def _add_grid(model, grid_object, check_for_duplicates = False):
    """Add grid object to list of shareable grids for this model."""

    if check_for_duplicates:
        for g in model.grid_list:
            if bu.matching_uuids(g.uuid, grid_object.uuid):
                return
    model.grid_list.append(grid_object)


def _grid_list_uuid_list(model):
    """Returns list of uuid's for the grid objects in the cached grid list."""

    uuid_list = []
    for grid in model.grid_list:
        uuid_list.append(grid.uuid)
    return uuid_list


def _grid_for_uuid_from_grid_list(model, uuid):
    """Returns the cached grid object matching the given uuid, if found in the grid list, otherwise None."""

    for grid in model.grid_list:
        if bu.matching_uuids(uuid, grid.uuid):
            return grid
    return None



"""Private common functions for derived model package."""

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.grid as grr
import resqpy.model as rq
import resqpy.olio.xml_et as rqet
import resqpy.property as rqp


def _displacement_properties(new_grid, old_grid):
    """Computes cell centre differences in x, y, & z, between old & new grids, and returns a collection of 3 properties."""

    displacement_collection = rqp.GridPropertyCollection()
    displacement_collection.set_grid(new_grid)
    old_grid.centre_point(cache_centre_array = True)
    new_grid.centre_point(cache_centre_array = True)
    displacement = new_grid.array_centre_point - old_grid.array_centre_point
    log.debug('displacement array shape: ' + str(displacement.shape))
    displacement_collection.x_array = displacement[..., 0].copy()
    displacement_collection.y_array = displacement[..., 1].copy()
    displacement_collection.z_array = displacement[..., 2].copy()
    # horizontal_displacement = np.sqrt(x_displacement * x_displacement  +  y_displacement * y_displacement)
    # todo: create prop collection to hold z_displacement and horizontal_displacement; add them to imported list
    xy_units = rqet.find_tag(new_grid.crs_root, 'ProjectedUom').text.lower()
    z_units = rqet.find_tag(new_grid.crs_root, 'VerticalUom').text.lower()
    # todo: could replace 3 displacement properties with a single points property
    displacement_collection.add_cached_array_to_imported_list(displacement_collection.x_array,
                                                              'easterly displacement from tilt',
                                                              'DX_DISPLACEMENT',
                                                              discrete = False,
                                                              uom = xy_units)
    displacement_collection.add_cached_array_to_imported_list(displacement_collection.y_array,
                                                              'northerly displacement from tilt',
                                                              'DY_DISPLACEMENT',
                                                              discrete = False,
                                                              uom = xy_units)
    displacement_collection.add_cached_array_to_imported_list(displacement_collection.z_array,
                                                              'vertical displacement from tilt',
                                                              'DZ_DISPLACEMENT',
                                                              discrete = False,
                                                              uom = z_units)
    return displacement_collection


def _pl(n, use_es = False):
    if n == 1:
        return ''
    elif use_es:
        return 'es'
    else:
        return 's'


def _prepare_simple_inheritance(grid, source_grid, inherit_properties, inherit_realization, inherit_all_realizations):
    collection = None
    if inherit_properties:
        source_collection = source_grid.extract_property_collection()
        if source_collection is not None:
            #  do not inherit the inactive property array by this mechanism
            collection = rqp.GridPropertyCollection()
            collection.set_grid(grid)
            collection.extend_imported_list_copying_properties_from_other_grid_collection(
                source_collection, realization = inherit_realization, copy_all_realizations = inherit_all_realizations)
    return collection


def _write_grid(epc_file,
                grid,
                ext_uuid = None,
                property_collection = None,
                grid_title = None,
                mode = 'a',
                geometry = True,
                time_series_uuid = None,
                string_lookup_uuid = None,
                extra_metadata = {}):
    """Append to or create epc and h5 files, with grid and optionally property collection.

    arguments:
       epc_file (string): name of existing epc file (if appending) or new epc file to be created (if writing)
       grid (grid.Grid object): the grid object to be written to the epc & h5 files
       ext_uuid (uuid.UUID object, optional): if present and the mode is 'a', the arrays are appended to
          the hdf5 file that has this uuid; if None or mode is 'w', the hdf5 file is determined automatically
       property_collection (property.GridPropertyCollection object, optional): if present, a collection of
          grid properties to write and relate to the grid
       grid_title (string): used as the citation title for the grid object
       mode (string, default 'a'): 'a' or 'w'; if 'a', epc_file should be an existing file which is extended
          (appended to) with the new grid and properties; if 'w', epc_file is created along with an h5 file
          and will be populated with the grid, crs and properties
       geometry (boolean, default True): if True, the grid object is included in the write; if False, only the
          property collection is written, in which case grid must be fully established with xml in place
       time_series_uuid (uuid.UUID, optional): the uuid of a time series object; required if property_collection
          contains any recurrent properties in its import list
       string_lookup_uuid (optional): if present, the uuid of the string table lookup which any non-continuous
          properties relate to (ie. they are all taken to be categorical); leave as None if discrete property
          objects are required rather than categorical
       extra_metadata (dict, optional): any items in this dictionary are added as extra metadata to any new
          properties

    returns:
       list of uuid.UUID, being the uuids of property parts added from the property_collection, if any

    note:
       this function is not usually called directly by application code
    """

    log.debug('write_grid(): epc_file: ' + str(epc_file) + '; mode: ' + str(mode) + '; grid extent: ' +
              str(grid.extent_kji))

    assert mode in ['a', 'w']
    assert geometry or mode == 'a', 'for now, a copy of the grid must be included with property collection'
    assert geometry or property_collection is not None, 'write_grid called without anything to write'

    if not epc_file.endswith('.epc'):
        epc_file += '.epc'

    is_regular = grr.is_regular_grid(grid.root)

    if not is_regular:
        grid.cache_all_geometry_arrays()
    working_model = grid.model

    if mode == 'a':  # append to existing model
        log.debug('write_grid(): re-using existing model object')
        model = working_model
        if ext_uuid is None:
            ext_uuid = model.h5_uuid()
    else:  # create new model with crs, grid and properties
        log.debug('write_grid(): creating new model object')
        model = rq.new_model(epc_file)
        ext_uuid = model.h5_uuid()
        crs_root = model.duplicate_node(grid.crs_root)
        grid.model = model
        grid.crs_root = crs_root
    log.debug('write_grid(): number of starting parts: ' + str(model.number_of_parts()))

    if grid.inactive is not None and geometry:
        inactive_count = np.count_nonzero(grid.inactive)
        if inactive_count == grid.inactive.size:
            log.warning('writing grid with all cells inactive')
        else:
            log.info(f'grid has {grid.inactive.size - inactive_count} active cells out of {grid.inactive.size}')
    collection = property_collection
    if collection is not None:
        log.debug('write_grid(): number of properties in collection: ' + str(collection.number_of_parts()))

    # append to hdf5 file using arrays cached in grid above
    hdf5_file = model.h5_file_name(uuid = ext_uuid, file_must_exist = (mode == 'a'))
    if geometry or collection is not None:
        log.debug('writing grid arrays to hdf5 file')
        grid.write_hdf5_from_caches(hdf5_file, mode = mode, geometry = geometry, imported_properties = collection)
        model.h5_release()
    if ext_uuid is None:
        ext_uuid = model.h5_uuid()

    # build xml for grid geometry
    if geometry:
        log.debug('building xml for grid object')
        ijk_node = grid.create_xml(ext_uuid, add_as_part = True, add_relationships = True, title = grid_title)
        assert ijk_node is not None, 'failed to create IjkGrid node in xml tree'
        if collection is not None:
            collection.set_grid(grid, grid_root = grid.root)
        grid.geometry_root = rqet.find_tag(ijk_node, 'Geometry')

    # add derived inactive array as part
    if collection is not None:
        prop_uuid_list = collection.create_xml_for_imported_list_and_add_parts_to_model(
            ext_uuid,
            support_uuid = grid.uuid,
            time_series_uuid = time_series_uuid,
            string_lookup_uuid = string_lookup_uuid,
            extra_metadata = extra_metadata)
    else:
        prop_uuid_list = []

    # set grid related short cuts in Model object
    if geometry:
        model.grid_list.append(grid)

    log.debug('write_grid(): number of finishing parts: ' + str(model.number_of_parts()))

    # store new version of model
    log.info('writing model xml data in epc file: ' + epc_file)
    model.store_epc(epc_file)

    return prop_uuid_list


def _establish_model_and_source_grid(epc_file, source_grid):
    assert epc_file or source_grid is not None, 'neither epc file name nor source grid supplied'
    if epc_file:
        model = rq.Model(epc_file)
        if source_grid is None:
            source_grid = model.grid()  # requires there to be exactly one grid in model (or one named ROOT)
    else:
        model = source_grid.model
    return model, source_grid



import os
import numpy as np
import pandas as pd

import resqpy.olio.xml_et as rqet
import resqpy.well
from resqpy.well.well_utils import load_hdf5_array, extract_xyz, find_entry_and_exit, _as_optional_array, _pl, well_names_in_cellio_file
from resqpy.grid import RegularGrid


def test_load_hdf5_array(example_model_with_well):

    # --------- Arrange ----------
    # Create a Deviation Survey object in memory

    # Load example model from a fixture
    model, well_interp, datum, traj = example_model_with_well

    # Create a survey
    data = dict(
        title = 'Majestic Umlaut ö',
        originator = 'Thor, god of sparkles',
        md_uom = 'ft',
        angle_uom = 'rad',
        is_final = True,
    )
    array_data = dict(
        measured_depths = np.array([1, 2, 3], dtype = float) + 1000.0,
        azimuths = np.array([4, 5, 6], dtype = float),
        inclinations = np.array([7, 8, 9], dtype = float),
        first_station = np.array([0, -1, 999], dtype = float),
    )

    survey = resqpy.well.DeviationSurvey(
        parent_model = model,
        represented_interp = well_interp,
        md_datum = datum,
        **data,
        **array_data,
    )

    # ----------- Act ---------
    survey.write_hdf5()
    survey.create_xml()
    mds_node = rqet.find_tag(survey.root, 'Mds', must_exist = True)
    # the only purpose of the call is to ensure that the array is cached in memory so should return None
    # a copy of the whole array is cached in memory as an attribute of the object
    expected_result = load_hdf5_array(survey, mds_node, 'measured_depths')

    # -------- Assert ---------
    np.testing.assert_equal(survey.__dict__['measured_depths'], array_data['measured_depths'])
    assert expected_result is None


def test_extract_xyz(example_model_with_well):

    # --------- Arrange ----------
    # Create a Deviation Survey object in memory

    # Load example model from a fixture
    model, well_interp, datum, traj = example_model_with_well

    # Create a survey
    data = dict(
        title = 'Majestic Umlaut ö',
        originator = 'Thor, god of sparkles',
        md_uom = 'ft',
        angle_uom = 'rad',
        is_final = True,
    )
    array_data = dict(
        measured_depths = np.array([1, 2, 3], dtype = float) + 1000.0,
        azimuths = np.array([4, 5, 6], dtype = float),
        inclinations = np.array([7, 8, 9], dtype = float),
        first_station = np.array([0, -1, 999], dtype = float),
    )

    survey = resqpy.well.DeviationSurvey(
        parent_model = model,
        represented_interp = well_interp,
        md_datum = datum,
        **data,
        **array_data,
    )

    # ----------- Act ---------
    survey.write_hdf5()
    survey.create_xml()
    first_station_node = rqet.find_tag(survey.root, 'FirstStationLocation', must_exist = True)
    first_station_xyz = extract_xyz(first_station_node)

    # ----------- Assert ---------
    np.testing.assert_equal(first_station_xyz, array_data['first_station'])


def test_find_entry_and_exit(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (5, 3, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    well_name = 'DOGLEG'
    bw = resqpy.well.BlockedWell(model, well_name = well_name, use_face_centres = True, add_wellspec_properties = True)
    # populate empty blocked well object for a 'vertical' well in the given column
    bw.set_for_column(well_name = well_name, grid = grid, col_ji0 = (1, 1))
    cp = grid.corner_points(cell_kji0 = (1, 1, 1))
    cell_centre = np.mean(cp, axis = (0, 1, 2))
    entry_xyz = np.array([25, -25, 125])
    exit_xyz = np.array([75, -50, 175])
    entry_vector = 100.0 * (entry_xyz - cell_centre)
    exit_vector = 100.0 * (exit_xyz - cell_centre)

    # --------- Act ----------
    (entry_axis, entry_polarity, entry_xyz, exit_axis, exit_polarity, exit_xyz) =\
        find_entry_and_exit(cp=cp, entry_vector=entry_vector, exit_vector=exit_vector, well_name=well_name)

    # --------- Assert ----------
    # TODO: less trivial assertion statement, understand the math behind the calcs
    for result in (entry_axis, entry_polarity, entry_xyz, exit_axis, exit_polarity, exit_xyz):
        assert result is not None


def test_as_optional_array():

    # --------- Arrange ----------
    to_test = [[1, 2, 3], None]

    # --------- Act ----------
    result1, result2 = _as_optional_array(to_test[0]), _as_optional_array(to_test[1])

    # --------- Assert ----------
    np.testing.assert_equal(result1, np.array([1, 2, 3]))
    assert result2 is None


def test_pl():

    # --------- Arrange ----------
    to_test = [1, 'Banoffee']

    # --------- Act ----------
    result1, result2, result3 = _pl(i = to_test[0]), _pl(i = to_test[1]), _pl(i = to_test[1], e = True)

    # --------- Assert ----------
    assert result1 == ''
    assert result2 == 's'
    assert result3 == 'es'


def test_well_names_in_cellio_file(tmp_path):

    # --------- Arrange ----------
    well_name = 'Banoffee'
    cellio_file = os.path.join(tmp_path, 'cellio.dat')
    source_df = pd.DataFrame(
        [[1, 1, 1, 25, 25, 0, 26, 26, 1, 120, 0.12], [2, 2, 1, 26, -26, 126, 27, -27, 127, 117, 0.20],
         [2, 3, 1, 27, -27, 127, 28, -28, 128, 135, 0.15]],
        columns = [
            'i_index unit1 scale1', 'j_index unit1 scale1', 'k_index unit1 scale1', 'x_in unit1 scale1',
            'y_in unit1 scale1', 'z_in unit1 scale1', 'x_out unit1 scale1', 'y_out unit1 scale1', 'z_out unit1 scale1',
            'Perm unit1 scale1', 'Poro unit1 scale1'
        ])

    with open(cellio_file, 'w') as fp:
        fp.write('1.0\n')
        fp.write('Undefined\n')
        fp.write(f'{well_name} terrible day\n')
        fp.write('11\n')
        for col in source_df.columns:
            fp.write(f' {col}\n')
        for row_index in range(len(source_df)):
            row = source_df.iloc[row_index]
            for col in source_df.columns:
                fp.write(f' {int(row[col])}')
            fp.write('\n')

    # --------- Arrange ----------
    well_list = well_names_in_cellio_file(cellio_file = cellio_file)

    # --------- Assert ----------
    assert len(well_list) == 1
    assert set(well_list) == {well_name}



import numpy as np
from numpy.testing import assert_array_almost_equal
import os
import resqpy.grid
import resqpy.grid_surface as rqgs
import resqpy.lines as rql
import resqpy.model as rq
import resqpy.olio.uuid as bu
import resqpy.organize
import resqpy.surface

import pytest

# Integration tests for surface classes


def test_surface(tmp_model):
    # Set up a Surface
    title = 'Mountbatten'
    model = tmp_model
    surf = resqpy.surface.Surface(parent_model = model, title = title)
    surf.create_xml()

    # Add a interpretation
    assert surf.represented_interpretation_root is None
    surf.create_interpretation_and_feature(kind = 'fault')
    assert surf.represented_interpretation_root is not None

    # Check fault can be loaded in again
    model.store_epc()
    fault_interp = resqpy.organize.FaultInterpretation(model, uuid = surf.represented_interpretation_uuid)
    fault_feature = resqpy.organize.TectonicBoundaryFeature(model, uuid = fault_interp.tectonic_boundary_feature.uuid)

    # Check title matches expected title
    assert fault_feature.feature_name == title


def test_faces_for_surface(tmp_model):
    crs = resqpy.crs.Crs(tmp_model)
    crs.create_xml()
    grid = resqpy.grid.RegularGrid(tmp_model, extent_kji = (3, 3, 3), crs_uuid = crs.uuid, set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    # todo: create sloping planar surface
    # call find faces for each of 3 different methods
    points = np.zeros((2, 2, 3))
    points[1, :, 1] = 3.0
    points[:, 1, 0] = 3.0
    points[:, 1, 2] = 3.0
    points[:, :, 2] += 0.25
    triangles = np.zeros((2, 3), dtype = int)
    triangles[0] = (0, 1, 2)
    triangles[1] = (3, 1, 2)
    surf = resqpy.surface.Surface(tmp_model, crs_uuid = crs.uuid)
    surf.set_from_triangles_and_points(triangles, points.reshape((-1, 3)))
    assert surf is not None
    for mode in ['staffa', 'regular', 'auto']:
        gcs = rqgs.find_faces_to_represent_surface(grid, surf, name = mode, mode = mode)
        assert gcs is not None
        assert gcs.count == 12
        cip = set([tuple(pair) for pair in gcs.cell_index_pairs])
        expected_cip = grid.natural_cell_indices(
            np.array([[[0, 0, 0], [1, 0, 0]], [[0, 1, 0], [1, 1, 0]], [[0, 2, 0], [1, 2, 0]], [[1, 0, 0], [1, 0, 1]],
                      [[1, 1, 0], [1, 1, 1]], [[1, 2, 0], [1, 2, 1]], [[1, 0, 1], [2, 0, 1]], [[1, 1, 1], [2, 1, 1]],
                      [[1, 2, 1], [2, 2, 1]], [[2, 0, 1], [2, 0, 2]], [[2, 1, 1], [2, 1, 2]], [[2, 2, 1], [2, 2, 2]]],
                     dtype = int))
        e_cip = set([tuple(pair) for pair in expected_cip])
        assert cip == e_cip  # note: this assumes lower cell index is first, which happens to be true
        # todo: check face indices
        gcs.write_hdf5()
        gcs.create_xml()
        assert bu.matching_uuids(
            tmp_model.uuid(obj_type = 'GridConnectionSetRepresentation', multiple_handling = 'newest'), gcs.uuid)


def test_delaunay_triangulation(example_model_and_crs):
    model, crs = example_model_and_crs

    # number of random points to use
    n = 20

    # create a set of random points
    x = np.random.random(n) * 1000.0
    y = np.random.random(n) * 1000.0
    z = np.random.random(n)  #  note: triangulation does not use z values
    p = np.stack((x, y, z), axis = -1)

    # make a PointSet object
    ps = resqpy.surface.PointSet(model, crs_uuid = crs.uuid, points_array = p, title = 'random points in square')

    # make another PointSet as random points within a closed polyline
    vertices = np.array(
        ((50.0, 99.0, 13.0), (85.0, 60.0, 17.5), (62.7, 11.0, 10.0), (33.3, 15.3, 19.2), (12.8, 57.8, 15.0)))
    polygon = rql.Polyline(model, set_crs = crs.uuid, set_bool = True, set_coord = vertices, title = 'the pentagon')
    polygon.write_hdf5()
    polygon.create_xml()
    ps2 = resqpy.surface.PointSet(model,
                                  crs_uuid = crs.uuid,
                                  polyline = polygon,
                                  random_point_count = n,
                                  title = 'random points in polygon')

    # process the point sets into triangulated surfaces
    for point_set in (ps, ps2):
        point_set.write_hdf5()
        point_set.create_xml()
        surf = resqpy.surface.Surface(model, point_set = point_set, title = 'surface from ' + str(point_set.title))
        assert surf is not None
        surf.write_hdf5()
        surf.create_xml()
        # check that coordinate range of points looks okay
        triangles, points = surf.triangles_and_points()
        assert len(points) == n
        original_points = point_set.full_array_ref()
        assert_array_almost_equal(np.nanmin(original_points, axis = 0), np.nanmin(points, axis = 0))
        assert_array_almost_equal(np.nanmax(original_points, axis = 0), np.nanmax(points, axis = 0))


@pytest.mark.parametrize('mesh_file,mesh_format,firstval', [('Surface_roxartext.txt', 'rms', 0.4229),
                                                            ('Surface_roxartext.txt', 'roxar', 0.4229),
                                                            ('Surface_zmap.dat', 'zmap', 0.4648)])
def test_surface_from_mesh_file(example_model_and_crs, test_data_path, mesh_file, mesh_format, firstval):
    # Arrange
    model, crs = example_model_and_crs
    in_file = test_data_path / mesh_file

    # Act
    surface = resqpy.surface.Surface(parent_model = model, mesh_file = in_file, mesh_format = mesh_format)

    # Assert
    assert surface is not None
    assert surface.patch_list[0].triangle_count == 12
    assert surface.patch_list[0].points[0][2] == firstval


def test_surface_from_tsurf_file(example_model_and_crs, test_data_path):
    # Arrange
    model, crs = example_model_and_crs
    in_file = test_data_path / 'Surface_tsurf.txt'

    # Act
    surface = resqpy.surface.Surface(parent_model = model, tsurf_file = in_file, title = 'horizon')
    surface.create_interpretation_and_feature(kind = 'horizon')

    # Assert
    assert surface is not None
    assert surface.represented_interpretation_root is not None
    assert surface.patch_list[0].triangle_count == 12
    assert surface.patch_list[0].points[0][2] == 0.4228516


def test_regular_mesh(example_model_and_crs):
    model, crs = example_model_and_crs

    #  number of points in mesh, origin spacing
    ni = 7
    nj = 5
    origin = (409000.0, 1605000.0, 0.0)
    di = dj = 50.0

    # create some random depths
    z = (np.random.random(ni * nj) * 20.0 + 1000.0).reshape((nj, ni))

    # make a regular mesh representation
    mesh = resqpy.surface.Mesh(model,
                               crs_uuid = crs.uuid,
                               mesh_flavour = 'reg&z',
                               ni = ni,
                               nj = nj,
                               origin = origin,
                               dxyz_dij = np.array([[di, 0.0, 0.0], [0.0, dj, 0.0]]),
                               z_values = z,
                               title = 'random mesh',
                               originator = 'Andy',
                               extra_metadata = {'testing mode': 'automated'})
    assert mesh is not None
    mesh.write_hdf5()
    mesh.create_xml()
    mesh_uuid = mesh.uuid

    # fully write model to disc
    model.store_epc()
    epc = model.epc_file

    # re-open model and check the mesh object is there
    model = rq.Model(epc)
    assert bu.matching_uuids(model.uuid(obj_type = 'Grid2dRepresentation', title = 'random mesh'), mesh_uuid)

    # establish a resqpy Mesh from the object in the RESQML dataset
    persistent_mesh = resqpy.surface.Mesh(model, uuid = mesh_uuid)

    # check some of the metadata
    assert persistent_mesh.ni == ni and persistent_mesh.nj == nj
    assert persistent_mesh.flavour == 'reg&z'
    assert_array_almost_equal(np.array(persistent_mesh.regular_origin), np.array(origin))
    assert_array_almost_equal(np.array(persistent_mesh.regular_dxyz_dij), np.array([[di, 0.0, 0.0], [0.0, dj, 0.0]]))

    # check a fully expanded version of the points
    assert_array_almost_equal(persistent_mesh.full_array_ref(), mesh.full_array_ref())

    # check that we can build a Surface from the Mesh
    surf = persistent_mesh.surface(quad_triangles = True)
    assert surf is not None

    # do some basic checks that the surface looks consistent with the mesh
    t, p = surf.triangles_and_points()
    assert len(p) == (ni * nj) + ((ni - 1) * (nj - 1))  # quad triangles mode introduces the extra points
    assert len(t) == 4 * (ni - 1) * (nj - 1)
    assert_array_almost_equal(np.min(p, axis = 0), np.min(persistent_mesh.full_array_ref().reshape(-1, 3), axis = 0))
    assert_array_almost_equal(np.max(p, axis = 0), np.max(persistent_mesh.full_array_ref().reshape(-1, 3), axis = 0))
    assert len(surf.distinct_edges()) == 6 * (ni - 1) * (nj - 1) + (ni - 1) + (nj - 1)


# @pytest.mark.skip(reason = "Bug in Mesh for ref&z flavou needs fixing first")
def test_refandz_mesh(example_model_and_crs):
    model, crs = example_model_and_crs

    # number of points in mesh, origin spacing
    ni = 7
    nj = 5
    origin = (409000.0, 1605000.0, 0.0)
    di = dj = 50.0

    # create some random depths
    z = (np.random.random(ni * nj) * 20.0 + 1000.0).reshape((nj, ni))
    z_values = z * 2.5

    # make a regular mesh representation
    support = resqpy.surface.Mesh(model,
                                  crs_uuid = crs.uuid,
                                  mesh_flavour = 'regular',
                                  ni = ni,
                                  nj = nj,
                                  origin = origin,
                                  dxyz_dij = np.array([[di, 0.0, 0.0], [0.0, dj, 0.0]]),
                                  z_values = z,
                                  title = 'random regular mesh',
                                  originator = 'Emma',
                                  extra_metadata = {'testing mode': 'automated'})
    assert support is not None
    support.write_hdf5()
    support.create_xml()
    support_uuid = support.uuid

    # model.store_epc()
    # model = rq.Model(model.epc_file)

    refz = resqpy.surface.Mesh(model,
                               crs_uuid = crs.uuid,
                               z_values = z_values,
                               ni = ni,
                               nj = nj,
                               z_supporting_mesh_uuid = support_uuid,
                               title = 'random refz mesh',
                               originator = 'Emma',
                               extra_metadata = {'testing mode': 'automated'})

    assert refz is not None
    refz.write_hdf5()
    refz.create_xml()
    refz_uuid = refz.uuid

    # fully write model to disc
    model.store_epc()

    # re-open model and check the mesh object is there
    reload = rq.Model(model.epc_file)

    assert bu.matching_uuids(reload.uuid(obj_type = 'Grid2dRepresentation', title = 'random refz mesh'), refz_uuid)

    # establish a resqpy Mesh from the object in the RESQML dataset
    reload_refzmesh = resqpy.surface.Mesh(reload, uuid = refz_uuid)

    # check some of the metadata
    assert reload_refzmesh.ni == ni and reload_refzmesh.nj == nj
    assert reload_refzmesh.flavour == 'ref&z'
    assert bu.matching_uuids(reload_refzmesh.ref_uuid, support_uuid)

    # check a fully expanded version of the points
    assert_array_almost_equal(reload_refzmesh.full_array_ref(), refz.full_array_ref())


def test_explicit_mesh(example_model_and_crs):
    model, crs = example_model_and_crs

    # create some random x,y,z values
    x = (np.random.random(20) * 10000)
    y = (np.random.random(20) * 10000)
    z = (np.random.random(20) * 500 + 2000)
    # make a 20x20x3 array of values to test
    array = np.array([x, y, z]).T
    xyz_values = np.array(
        [np.multiply(array,
                     np.array([1 + val / 100, 1. + val / 100, 1 + val / 100]).T) for val in np.arange(0, 20)])

    # make an explicit mesh representation
    mesh = resqpy.surface.Mesh(model,
                               crs_uuid = crs.uuid,
                               mesh_flavour = 'explicit',
                               xyz_values = xyz_values,
                               title = 'random explicit mesh',
                               originator = 'Emma',
                               extra_metadata = {'testing mode': 'automated'})
    assert mesh is not None
    mesh.write_hdf5()
    mesh.create_xml()
    mesh_uuid = mesh.uuid

    # fully write model to disc
    model.store_epc()
    epc = model.epc_file

    # re-open model and check the mesh object is there
    model = rq.Model(epc)
    assert bu.matching_uuids(model.uuid(obj_type = 'Grid2dRepresentation', title = 'random explicit mesh'), mesh_uuid)

    # establish a resqpy Mesh from the object in the RESQML dataset
    persistent_mesh = resqpy.surface.Mesh(model, uuid = mesh_uuid)

    # check some of the metadata
    assert persistent_mesh.flavour == 'explicit'
    assert persistent_mesh.ni == len(x)
    assert persistent_mesh.nj == len(y)

    # check a fully expanded version of the points
    assert_array_almost_equal(persistent_mesh.full_array_ref(), mesh.full_array_ref())


@pytest.mark.parametrize('flavour,infile,filetype', [('explicit', 'Surface_roxartext.txt', 'roxar'),
                                                     ('explicit', 'Surface_roxartext.txt', 'rms'),
                                                     ('explicit', 'Surface_zmap.dat', 'zmap'),
                                                     ('regular', 'Surface_roxartext.txt', 'roxar'),
                                                     ('regular', 'Surface_roxartext.txt', 'rms'),
                                                     ('regular', 'Surface_zmap.dat', 'zmap'),
                                                     ('reg&z', 'Surface_roxartext.txt', 'roxar'),
                                                     ('reg&z', 'Surface_roxartext.txt', 'rms'),
                                                     ('reg&z', 'Surface_zmap.dat', 'zmap')])
def test_mesh_file(example_model_and_crs, test_data_path, flavour, infile, filetype):
    model, crs = example_model_and_crs

    mesh_file = test_data_path / infile

    # make an explicit mesh representation
    mesh = resqpy.surface.Mesh(model,
                               crs_uuid = crs.uuid,
                               mesh_flavour = flavour,
                               mesh_file = mesh_file,
                               mesh_format = filetype,
                               title = 'mesh from file',
                               originator = 'Emma',
                               extra_metadata = {'testing mode': 'automated'})
    assert mesh is not None
    mesh.write_hdf5()
    mesh.create_xml()
    mesh_uuid = mesh.uuid

    # fully write model to disc
    model.store_epc()
    epc = model.epc_file

    # re-open model and check the mesh object is there
    model = rq.Model(epc)
    assert bu.matching_uuids(model.uuid(obj_type = 'Grid2dRepresentation', title = 'mesh from file'), mesh_uuid)

    # establish a resqpy Mesh from the object in the RESQML dataset
    persistent_mesh = resqpy.surface.Mesh(model, uuid = mesh_uuid)

    # check some of the metadata
    assert persistent_mesh.flavour == flavour

    # check a fully expanded version of the points
    if flavour != 'regular':
        assert_array_almost_equal(persistent_mesh.full_array_ref(), mesh.full_array_ref())


def test_pointset_from_array(example_model_and_crs):
    model, crs = example_model_and_crs

    # create some random x,y,z values
    x = (np.random.random(50) * 10000)
    y = (np.random.random(50) * 10000)
    z = (np.random.random(50) * 500 + 2000)

    # make a pointset representation
    points = resqpy.surface.PointSet(model,
                                     crs_uuid = crs.uuid,
                                     points_array = np.array([x, y, z]).T,
                                     title = 'random points',
                                     originator = 'Emma',
                                     extra_metadata = {'testing mode': 'automated'})
    assert points is not None
    points.create_interpretation_and_feature(kind = 'horizon')
    points.write_hdf5()
    points.create_xml()
    points_uuid = points.uuid

    # fully write model to disc
    model.store_epc()
    epc = model.epc_file

    # re-open model and check the points object is there
    model = rq.Model(epc)
    assert bu.matching_uuids(model.uuid(obj_type = 'PointSetRepresentation', title = 'random points'), points_uuid)

    # establish a resqpy Pointset from the object in the RESQML dataset
    saved_points = resqpy.surface.PointSet(model, uuid = points_uuid)

    # check a fully expanded version of the points
    assert_array_almost_equal(saved_points.full_array_ref(), points.full_array_ref())
    assert saved_points.represented_interpretation_root is not None


def test_pointset_from_2d_array(example_model_and_crs):
    model, crs = example_model_and_crs

    # create some random x,y values
    x = (np.random.random(50) * 10000)
    y = (np.random.random(50) * 10000)

    # make a pointset representation
    points = resqpy.surface.PointSet(model,
                                     crs_uuid = crs.uuid,
                                     points_array = np.array([x, y]).T,
                                     title = 'random 2d points',
                                     originator = 'Emma',
                                     extra_metadata = {'testing mode': 'automated'})
    assert points is not None
    points.write_hdf5()
    points.create_xml()
    points_uuid = points.uuid

    # fully write model to disc
    model.store_epc()
    epc = model.epc_file

    # re-open model and check the points object is there
    model = rq.Model(epc)
    assert bu.matching_uuids(model.uuid(obj_type = 'PointSetRepresentation', title = 'random 2d points'), points_uuid)

    # establish a resqpy Pointset from the object in the RESQML dataset
    saved_points = resqpy.surface.PointSet(model, uuid = points_uuid)

    # check a fully expanded version of the points
    assert_array_almost_equal(saved_points.full_array_ref(), points.full_array_ref())


def test_pointset_from_array_multipatch(example_model_and_crs):
    model, crs = example_model_and_crs

    # create some random x,y values
    x = (np.random.random(50) * 10000)
    y = (np.random.random(50) * 10000)

    x2 = (np.random.random(50) * 10000)
    y2 = (np.random.random(50) * 10000)

    # make a pointset representation
    points = resqpy.surface.PointSet(model,
                                     crs_uuid = crs.uuid,
                                     points_array = np.array([x, y]).T,
                                     title = 'random 2d points',
                                     originator = 'Emma',
                                     extra_metadata = {'testing mode': 'automated'})
    assert points is not None
    points.add_patch(points_array = np.array([x2, y2]).T)
    points.write_hdf5()
    points.create_xml()
    points_uuid = points.uuid

    # fully write model to disc
    model.store_epc()
    epc = model.epc_file

    # re-open model and check the points object is there
    model = rq.Model(epc)
    assert bu.matching_uuids(model.uuid(obj_type = 'PointSetRepresentation', title = 'random 2d points'), points_uuid)

    # establish a resqpy Pointset from the object in the RESQML dataset
    saved_points = resqpy.surface.PointSet(model, uuid = points_uuid)

    # check a fully expanded version of the points
    assert_array_almost_equal(saved_points.full_array_ref(), points.full_array_ref())
    assert_array_almost_equal(saved_points.single_patch_array_ref(0), np.array([x, y, [0] * 50]).T)
    assert_array_almost_equal(saved_points.single_patch_array_ref(1), np.array([x2, y2, [0] * 50]).T)


def test_pointset_from_charisma(example_model_and_crs, test_data_path, tmp_path):
    # Set up a PointSet and save to resqml file
    model, crs = example_model_and_crs

    charisma_file = test_data_path / "Charisma_points.txt"
    points = resqpy.surface.PointSet(parent_model = model, charisma_file = str(charisma_file), crs_uuid = crs.uuid)
    points.write_hdf5()
    points.create_xml()
    model.store_epc()

    # Test reload from resqml
    model = rq.Model(epc_file = model.epc_file)
    reload = resqpy.surface.PointSet(parent_model = model, uuid = points.uuid)

    assert reload.title == str(charisma_file)

    coords = reload.full_array_ref()
    assert_array_almost_equal(coords[0], np.array([420691.19624, 6292314.22044, 2799.05591]))

    assert coords.shape == (15, 3), f'Expected shape (15,3), not {coords.shape}'

    # Test write back to file
    out_file = str(tmp_path / "Charisma_points_out.txt")
    reload.convert_to_charisma(out_file)

    assert os.path.exists(out_file)
    with open(out_file, 'r') as f:
        line = f.readline()

    assert line == 'INLINE :\t1 XLINE :\t1\t420691.19624\t6292314.22044\t2799.05591\n', 'Output Charisma file does not look as expected'


def test_pointset_from_irap(example_model_and_crs, test_data_path, tmp_path):
    # Set up a PointSet and save to resqml file
    model, crs = example_model_and_crs
    irap_file = test_data_path / "IRAP_points.txt"
    points = resqpy.surface.PointSet(parent_model = model, irap_file = str(irap_file), crs_uuid = crs.uuid)
    points.write_hdf5()
    points.create_xml()
    model.store_epc()

    # Test reload from resqml
    model = rq.Model(epc_file = model.epc_file)
    reload = resqpy.surface.PointSet(parent_model = model, uuid = points.uuid)

    assert reload.title == str(irap_file)

    coords = reload.full_array_ref()
    assert_array_almost_equal(coords[0], np.array([429450.658333, 6296954.224574, 2403.837646]))

    assert coords.shape == (9, 3), f'Expected shape (9,3), not {coords.shape}'

    # Test write back to file
    out_file = str(tmp_path / "IRAP_points_out.txt")
    reload.convert_to_irap(out_file)

    assert os.path.exists(out_file)
    with open(out_file, 'r') as f:
        line = f.readline()

    assert line == '429450.658333 6296954.224574 2403.837646\n', 'Output IRAP file does not look as expected'


def test_pointset_from_polyline(example_model_and_crs):
    # Set up a PolyLine and save to resqml file
    model, crs = example_model_and_crs

    coords = np.array([[1, 2, 3, 4, 5], [6, 7, 8, 9, 10], [12, 13, 14, 15, 16]]).T
    lines = resqpy.lines.Polyline(parent_model = model,
                                  set_coord = coords,
                                  title = 'Polylines',
                                  set_crs = crs.uuid,
                                  set_bool = False)
    lines.write_hdf5()
    lines.create_xml()
    model.store_epc()

    # Reload the model, and generate pointset using the polyline
    model = rq.Model(epc_file = model.epc_file)
    reload_lines = resqpy.lines.Polyline(parent_model = model, uuid = lines.uuid)

    points = resqpy.surface.PointSet(parent_model = model, polyline = reload_lines, crs_uuid = crs.uuid)
    points.write_hdf5()
    points.create_xml()
    model.store_epc()

    # Reload the model, and ensure the coordinates are as expected
    model = rq.Model(epc_file = model.epc_file)
    reload = resqpy.surface.PointSet(parent_model = model, uuid = points.uuid)
    assert_array_almost_equal(reload.full_array_ref(), coords)


def test_pointset_from_polylineset(example_model_and_crs):
    # Set up a PolyLine and save to resqml file
    model, crs = example_model_and_crs

    coords1 = np.array([[1, 2, 3, 4, 5], [6, 7, 8, 9, 10], [12, 13, 14, 15, 16]]).T
    coords2 = np.array([[11, 12, 13, 14, 15], [16, 17, 18, 19, 110], [112, 113, 114, 115, 116]]).T

    line1 = resqpy.lines.Polyline(parent_model = model,
                                  title = 'Polyline1',
                                  set_coord = coords1,
                                  set_crs = crs.uuid,
                                  set_bool = False)

    line2 = resqpy.lines.Polyline(parent_model = model,
                                  title = 'Polyline2',
                                  set_coord = coords2,
                                  set_crs = crs.uuid,
                                  set_bool = False)

    lines = resqpy.lines.PolylineSet(parent_model = model, title = 'Polylines', polylines = [line1, line2])
    lines.write_hdf5()
    lines.create_xml()
    model.store_epc()

    # Reload the model, and generate pointset using the polylineset
    model = rq.Model(epc_file = model.epc_file)
    reload_lines = resqpy.lines.PolylineSet(parent_model = model, uuid = lines.uuid)

    points = resqpy.surface.PointSet(parent_model = model, polyset = reload_lines, crs_uuid = crs.uuid)
    points.write_hdf5()
    points.create_xml()
    model.store_epc()

    # Reload the model, and ensure the coordinates are as expected
    model = rq.Model(epc_file = model.epc_file)
    reload = resqpy.surface.PointSet(parent_model = model, uuid = points.uuid)

    assert_array_almost_equal(reload.full_array_ref(), np.concatenate((coords1, coords2), axis = 0))


def test_tripatch_set_to_triangle(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    corners = np.array([[0, 0, 0], [0, 1, 1], [1, 0, 1]])

    # Act
    tripatch = resqpy.surface.TriangulatedPatch(parent_model = model)
    tripatch.set_to_triangle(corners)

    # Assert
    assert tripatch is not None
    assert_array_almost_equal(tripatch.triangles, np.array([[0, 1, 2]]))
    assert_array_almost_equal(tripatch.points, corners)


def test_tripatch_verticalscale(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    corners = np.array([[0, 0, 0], [0, 1, 1], [1, 0, 1]])

    # Act
    tripatch = resqpy.surface.TriangulatedPatch(parent_model = model)
    tripatch.set_to_triangle(corners)

    # Assert
    assert tripatch is not None
    # Scale without a reference depth initially
    tripatch.vertical_rescale_points(0, 10)
    assert_array_almost_equal(tripatch.points[:, 2], np.array([0, 10, 10]))
    # Scale with a reference depth
    tripatch.vertical_rescale_points(5, 2)
    assert_array_almost_equal(tripatch.points[:, 2], np.array([-5, 15, 15]))


def test_tripatch_set_from_irregularmesh(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    mesh_xyz = np.array([[[0, 0, 1], [1, 0, 1]], [[0, 1, 1], [1, 1, 1]]])

    # Act
    tripatch = resqpy.surface.TriangulatedPatch(parent_model = model)
    tripatch.set_from_irregular_mesh(mesh_xyz = mesh_xyz, quad_triangles = False)

    # Assert
    assert tripatch is not None
    assert tripatch.node_count == 4
    assert tripatch.points.shape == (4, 3)
    assert_array_almost_equal(tripatch.points[0], mesh_xyz[0, 0])
    assert_array_almost_equal(tripatch.points[3], mesh_xyz[1, 1])
    assert_array_almost_equal(tripatch.triangles, np.array([[0, 1, 2], [3, 1, 2]]))


def test_tripatch_columnfromindex(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    mesh_xyz = np.array([[[0, 0, 1], [1, 0, 1], [2, 0, 2]], [[0, 1, 1], [1, 1, 1], [2, 1, 1]]])

    # Act
    tripatch = resqpy.surface.TriangulatedPatch(parent_model = model)
    tripatch.set_from_irregular_mesh(mesh_xyz = mesh_xyz, quad_triangles = False)
    assert tripatch is not None
    j0, i0 = tripatch.column_from_triangle_index(0)
    j1, i1 = tripatch.column_from_triangle_index(np.array([0, 3]))

    # Assert
    assert j0 == 0
    assert i0 == 0
    assert_array_almost_equal(j1, np.array([0, 0]))
    assert_array_almost_equal(i1, np.array([0, 1]))


def test_tripatch_columnfromindex_quad(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    mesh_xyz = np.array([[[0, 0, 1], [1, 0, 1], [2, 0, 2]], [[0, 1, 1], [1, 1, 1], [2, 1, 1]]])

    # Act
    tripatch = resqpy.surface.TriangulatedPatch(parent_model = model)
    tripatch.set_from_irregular_mesh(mesh_xyz = mesh_xyz, quad_triangles = True)
    assert tripatch is not None
    j0, i0 = tripatch.column_from_triangle_index(0)
    j1, i1 = tripatch.column_from_triangle_index(np.array([0, 7]))
    j2, i2 = tripatch.column_from_triangle_index(8)
    j3, i3 = tripatch.column_from_triangle_index(np.array([0, 8]))

    # Assert
    assert j0 == 0
    assert i0 == 0
    assert_array_almost_equal(j1, np.array([0, 0]))
    assert_array_almost_equal(i1, np.array([0, 1]))
    assert j2 is None
    assert i2 is None
    assert j3 is None
    assert i3 is None


def test_tripatch_set_from_irregularmesh_quad(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    mesh_xyz = np.array([[[0, 0, 1], [1, 0, 1]], [[0, 1, 1], [1, 1, 1]]])

    # Act
    tripatch = resqpy.surface.TriangulatedPatch(parent_model = model)
    tripatch.set_from_irregular_mesh(mesh_xyz = mesh_xyz, quad_triangles = True)

    # Assert
    assert tripatch is not None
    assert tripatch.node_count == 5
    assert tripatch.points.shape == (5, 3)
    assert_array_almost_equal(tripatch.points[0], mesh_xyz[0, 0])
    assert_array_almost_equal(tripatch.points[3], mesh_xyz[1, 1])
    assert_array_almost_equal(tripatch.triangles, np.array([[4, 0, 1], [4, 1, 3], [4, 3, 2], [4, 2, 0]]))


def test_tripatch_set_from_sparse(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    mesh_xyz = np.array([[[0, 0, 1], [1, 0, 1], [2, 0, 2]], [[0, 1, 1], [1, 1, np.nan], [2, 1, 1]],
                         [[0, 2, 1], [1, 2, 1], [2, 2, 1]]])

    # Act
    tripatch = resqpy.surface.TriangulatedPatch(parent_model = model)
    tripatch.set_from_sparse_mesh(mesh_xyz = mesh_xyz)

    # Assert
    assert tripatch is not None
    assert tripatch.node_count == 8
    assert tripatch.points.shape == (8, 3)
    assert_array_almost_equal(tripatch.points[0], mesh_xyz[0, 0])
    assert_array_almost_equal(tripatch.points[7], mesh_xyz[2, 2])
    assert_array_almost_equal(tripatch.triangles[0], np.array([1, 3, 0]))
    assert_array_almost_equal(tripatch.triangles[3], np.array([4, 6, 7]))


def test_tripatch_set_from_torn(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    mesh_xyz = np.array([[[[[0, 0, 1], [1, 0, 1]], [[0, 1, 1], [1, 1, np.nan]]],
                          [[[1, 0, 1], [2, 0, 1]], [[1, 1, np.nan], [2, 1, 1]]]]])

    # Act
    tripatch = resqpy.surface.TriangulatedPatch(parent_model = model)
    tripatch.set_from_torn_mesh(mesh_xyz = mesh_xyz)

    # Assert
    assert tripatch is not None
    assert tripatch.node_count == 8
    assert tripatch.points.shape == (8, 3)
    assert tripatch.ni == 2
    assert_array_almost_equal(tripatch.points[0], mesh_xyz[0, 0, 0, 0])
    assert_array_almost_equal(tripatch.points[2], mesh_xyz[0, 0, 1, 0])
    assert_array_almost_equal(tripatch.triangles, np.array([[0, 1, 2], [3, 1, 2], [4, 5, 6], [7, 5, 6]]))


def test_tripatch_set_from_torn_quad(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    mesh_xyz = np.array([[[[[0, 0, 1], [1, 0, 1]], [[0, 1, 1], [1, 1, np.nan]]],
                          [[[1, 0, 1], [2, 0, 1]], [[1, 1, np.nan], [2, 1, 1]]]]])

    # Act
    tripatch = resqpy.surface.TriangulatedPatch(parent_model = model)
    tripatch.set_from_torn_mesh(mesh_xyz = mesh_xyz, quad_triangles = True)

    # Assert
    assert tripatch is not None
    assert tripatch.node_count == 10
    assert tripatch.points.shape == (10, 3)
    assert tripatch.ni == 2
    assert_array_almost_equal(tripatch.points[0], mesh_xyz[0, 0, 0, 0])
    assert_array_almost_equal(tripatch.points[2], mesh_xyz[0, 0, 1, 0])
    assert_array_almost_equal(
        tripatch.triangles,
        np.array([[8, 0, 1], [8, 1, 3], [8, 3, 2], [8, 2, 0], [9, 4, 5], [9, 5, 7], [9, 7, 6], [9, 6, 4]]))


def test_tripatch_set_cellface_corp(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    cp = np.array([[[[0, 0, 0], [0, 1, 0]], [[1, 1, 0], [1, 0, 0]]], [[[0, 0, 1], [0, 1, 1]], [[1, 1, 1], [1, 0, 1]]]])

    # Act
    tripatch = resqpy.surface.TriangulatedPatch(parent_model = model)
    tripatch.set_to_cell_faces_from_corner_points(cp = cp)

    # Assert
    assert tripatch is not None
    assert tripatch.node_count == 14
    assert tripatch.points.shape == (14, 3)
    assert_array_almost_equal(tripatch.points[0], cp[0, 0, 0])
    assert_array_almost_equal(tripatch.points[2], cp[0, 1, 0])
    assert_array_almost_equal(tripatch.triangles[0], np.array([8, 0, 1]))
    assert_array_almost_equal(tripatch.triangles[-1], np.array([13, 3, 7]))


def test_tripatch_set_cellface_corp_quadfalse(example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    cp = np.array([[[[0, 0, 0], [0, 1, 0]], [[1, 1, 0], [1, 0, 0]]], [[[0, 0, 1], [0, 1, 1]], [[1, 1, 1], [1, 0, 1]]]])

    # Act
    tripatch = resqpy.surface.TriangulatedPatch(parent_model = model)
    tripatch.set_to_cell_faces_from_corner_points(cp = cp, quad_triangles = False)

    # Assert
    assert tripatch is not None
    assert tripatch.node_count == 8
    assert tripatch.points.shape == (8, 3)
    assert_array_almost_equal(tripatch.points[0], cp[0, 0, 0])
    assert_array_almost_equal(tripatch.points[2], cp[0, 1, 0])
    assert_array_almost_equal(tripatch.triangles[0], np.array([0, 3, 1]))
    assert_array_almost_equal(tripatch.triangles[-1], np.array([7, 1, 3]))



"""Generic small utility functions."""

version = '29th April 2021'

import numpy as np


def kji0_from_ijk1(index_ijk1):  # reverse order of indices and subtract 1
    """Returns index converted from simulator protocol to python protocol."""
    dims = index_ijk1.size
    result = np.zeros(dims, dtype = 'int')
    for d in range(dims):
        result[d] = index_ijk1[dims - d - 1] - 1
    return result


def ijk1_from_kji0(index_kji0):  # reverse order of indices and add 1
    """Returns index converted from python protocol to simulator protocol."""
    dims = index_kji0.size
    result = np.zeros(dims, dtype = 'int')
    for d in range(dims):
        result[d] = index_kji0[dims - d - 1] + 1
    return result


def extent_switch_ijk_kji(extent_in):  # reverse order of elements in extent
    """Returns equivalent grid extent switched either way between simulator and python protocols."""
    dims = extent_in.size
    result = np.zeros(dims, dtype = 'int')
    for d in range(dims):
        result[d] = extent_in[dims - d - 1]
    return result


def cell_count_from_extent(extent):
    """Returns the number of cells in a grid with the given extent."""
    result = 1
    for d in range(len(extent)):  # list, tuple or 1D numpy array
        result *= extent[d]
    return result


def string_ijk1_for_cell_kji0(cell_kji0):
    """Returns a string showing indices for a cell in simulator protocol, from data in python protocol."""
    return '[{:}, {:}, {:}]'.format(cell_kji0[2] + 1, cell_kji0[1] + 1, cell_kji0[0] + 1)


def string_ijk1_for_cell_ijk1(cell_ijk1):
    """Returns a string showing indices for a cell in simulator protocol, from data in simulator protocol."""
    return '[{:}, {:}, {:}]'.format(cell_ijk1[0], cell_ijk1[1], cell_ijk1[2])


def string_ijk_for_extent_kji(extent_kji):
    """Returns a string showing grid extent in simulator protocol, from data in python protocol."""
    return '[{:}, {:}, {:}]'.format(extent_kji[2], extent_kji[1], extent_kji[0])


def string_xyz(xyz):
    """Returns an xyz point as a string like (121234.56, 567890.12, 3456.789)"""

    return '({0:4.2f}, {1:4.2f}, {2:5.3f})'.format(xyz[0], xyz[1], xyz[2])


def horizon_float(k0, plus_or_minus):
    """Returns a floating point representation of k direction face index; eg.

    3.5 for face between layers 3 and 4.
    """
    result = float(k0)
    if plus_or_minus == '+':
        result += 0.5
    elif plus_or_minus == '-':
        result -= 0.5
    else:
        assert False
    return result



import numpy as np
from numpy.testing import assert_array_almost_equal
import resqpy.grid
import resqpy.organize
import resqpy.surface

import pytest

# Unit tests for surface/Pointset methods


def test_from_charisma_method(mocker, example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    charisma_file = "INLINE :  25701 XLINE :  23693    420691.19624   6292314.22044      2799.05591\nINLINE :  25701 XLINE :  23694    420680.15765   6292308.35532      2798.08496"
    open_mock = mocker.mock_open(read_data = charisma_file)
    mocker.patch("builtins.open", open_mock)
    test_path = "path/to/file"

    array = np.array([[420691.19624, 6292314.22044, 2799.05591], [420680.15765, 6292308.35532, 2798.08496]])

    patch_mock = mocker.MagicMock(name = 'add_patch')
    # mocker.patch(resqpy.surface.PointSet.add_patch, patch_mock)

    # Act
    pointset = resqpy.surface.PointSet(model, crs_uuid = crs.uuid)
    pointset.add_patch = patch_mock
    pointset.from_charisma("path/to/file")

    # Assert
    open_mock.assert_called_once_with(test_path, 'r')
    patch_mock.assert_called_once()
    assert_array_almost_equal(patch_mock.call_args[0][0], array)


def test_from_charisma_method_failure(mocker, example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    charisma_file = "INLINE :  25701 XLINE :  23693    420691.19624   6292314.22044      2799.05591\nINLINE :  25701 XLINE :  23694    420680.15765   6292308.35532      2798.08496"
    open_mock = mocker.mock_open(read_data = charisma_file)
    mocker.patch("builtins.open", open_mock)

    # Act
    with pytest.raises(AssertionError) as e_info:
        pointset = resqpy.surface.PointSet(model, charisma_file = "path/to/file")

    # Assert
    assert str(e_info.value) == 'crs uuid missing when establishing point set from charisma file'


def test_from_irap_method(mocker, example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    irap_file = "429450.658333 6296954.224574 2403.837646\n429444.793211 6296965.263155 2403.449707"
    open_mock = mocker.mock_open(read_data = irap_file)
    mocker.patch("builtins.open", open_mock)
    test_path = "path/to/file"

    array = np.array([[429450.658333, 6296954.224574, 2403.837646], [429444.793211, 6296965.263155, 2403.449707]])

    patch_mock = mocker.MagicMock(name = 'add_patch')

    # Act
    pointset = resqpy.surface.PointSet(model, crs_uuid = crs.uuid)
    pointset.add_patch = patch_mock
    pointset.from_irap("path/to/file")

    # Assert
    open_mock.assert_called_once_with(test_path, 'r')
    patch_mock.assert_called_once()
    assert_array_almost_equal(patch_mock.call_args[0][0], array)


def test_from_irap_method_failure(mocker, example_model_and_crs):
    # Arrange
    model, crs = example_model_and_crs
    irap_file = "429450.658333 6296954.224574 2403.837646\n429444.793211 6296965.263155 2403.449707"
    open_mock = mocker.mock_open(read_data = irap_file)
    mocker.patch("builtins.open", open_mock)

    # Act
    with pytest.raises(AssertionError) as e_info:
        pointset = resqpy.surface.PointSet(model, irap_file = "path/to/file")

    # Assert
    assert str(e_info.value) == 'crs uuid missing when establishing point set from irap file'


@pytest.mark.parametrize('closed,coords1,coords2,expected',
                         [(True, np.array([[2, 2, 2], [3, 3, 3], [2, 2, 2]]), np.array(
                             [[1, 1, 1]]), np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])),
                          (False, np.array([[2, 2, 2], [3, 3, 3], [2, 2, 2]]), np.array(
                              [[1, 1, 1]]), np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3], [2, 2, 2]]))])
def test_concat_polyset_points(closed, coords1, coords2, expected):
    # Act
    result = resqpy.surface.pointset.concat_polyset_points(closed, coords1, coords2)
    # Assert
    assert_array_almost_equal(result, expected)



"""volume.py: Functions to calculate volumes of hexahedral cells; assumes consistent length units."""

version = '31st August 2021'

import numpy as np


def _pyr(ra, ab, ac, ad, db):  # returns 6 * volume of one quad pyramid, defined by specific vectors
    return np.dot(ra, np.cross(db, ac)) + 0.5 * np.dot(ac, np.cross(ad, ab))


def pyramid_volume(apex, a, b, c, d, crs_is_right_handed = False):
    """Returns volume of a quadrilateral pyramid.

    arguments:
       apex (triple float): location of the apex of the pyramid
       a, b, c, d (each triple float): locations of corners of base of pyramid; clockwise viewed from apex
          for a left handed crs
       crs_is_right_handed (boolean, default False): set True if xyz axes of crs are right handed

    returns:
       float, being the volume of the pyramid; units are implied by crs units in use by the vertices
    """

    apex = np.array(apex)
    a = np.array(a)
    b = np.array(b)
    c = np.array(c)
    d = np.array(d)
    v = _pyr(a - apex, b - a, c - a, d - a, b - d) / 6.0
    if not crs_is_right_handed:
        v = -v
    return v


def tetrahedron_volume(a, b, c, d):
    """Returns volume of a tetrahedron.

    arguments:
       a, b, c, d (each triple float): locations of corners of tetrahedron
       crs_is_right_handed (boolean, default False): set True if xyz axes of crs are right handed

    returns:
       float, being the volume of the tetrahedron; units are implied by crs units in use by the vertices
    """

    a = np.array(a)
    b = np.array(b)
    c = np.array(c)
    d = np.array(d)

    return np.abs(np.dot(a - d, np.cross(b - d, c - d))) / 6.0


def tetra_cell_volume(cp, centre = None, off_hand = False):
    """Returns volume of single hexahedral cell with corner points cp of shape (2, 2, 2, 3); assumes bilinear faces."""

    if centre is None:
        centre = np.mean(cp.reshape((8, 3)), axis = 0)
    r0 = cp[0, 0, 0] - centre
    r1 = cp[1, 1, 1] - centre
    v = 0.0
    v += _pyr(r0, cp[0, 0, 1] - cp[0, 0, 0], cp[0, 1, 1] - cp[0, 0, 0], cp[0, 1, 0] - cp[0, 0, 0],
              cp[0, 1, 0] - cp[0, 0, 1])
    v += _pyr(r0, cp[1, 0, 0] - cp[0, 0, 0], cp[1, 0, 1] - cp[0, 0, 0], cp[0, 0, 1] - cp[0, 0, 0],
              cp[0, 0, 1] - cp[1, 0, 0])
    v += _pyr(r0, cp[0, 1, 0] - cp[0, 0, 0], cp[1, 1, 0] - cp[0, 0, 0], cp[1, 0, 0] - cp[0, 0, 0],
              cp[1, 0, 0] - cp[0, 1, 0])
    v += _pyr(r1, cp[1, 0, 1] - cp[1, 1, 1], cp[1, 0, 0] - cp[1, 1, 1], cp[1, 1, 0] - cp[1, 1, 1],
              cp[1, 1, 0] - cp[1, 0, 1])
    v += _pyr(r1, cp[1, 1, 0] - cp[1, 1, 1], cp[0, 1, 0] - cp[1, 1, 1], cp[0, 1, 1] - cp[1, 1, 1],
              cp[0, 1, 1] - cp[1, 1, 0])
    v += _pyr(r1, cp[0, 1, 1] - cp[1, 1, 1], cp[0, 0, 1] - cp[1, 1, 1], cp[1, 0, 1] - cp[1, 1, 1],
              cp[1, 0, 1] - cp[0, 1, 1])

    if off_hand:
        v = -v

    return v / 6.0


def tetra_volumes_slow(cp, centres = None, off_hand = False):
    """Returns volume array for all hexahedral cells assuming bilinear faces, using loop over cells."""

    # NB: deprecated, superceded by much faster function below
    # todo: handle NaNs
    # Pagoda style corner point data
    assert cp.ndim == 7

    flat = cp.reshape(-1, 2, 2, 2, 3)
    cells = flat.shape[0]
    if centres is None:
        centres = np.mean(flat, axis = (1, 2, 3))
    else:
        centres = centres.reshape((-1, 3))
    volumes = np.zeros(cells)
    for cell in range(cells):
        volumes[cell] = tetra_cell_volume(flat[cell], centre = centres[cell], off_hand = off_hand)
    return volumes.reshape(cp.shape[0:3])


def tetra_volumes(cp, centres = None, off_hand = False):
    """Returns volume array for all hexahedral cells assuming bilinear faces, using numpy operations.

    arguments:
       cp (7D numpy array of floats): cell corner point data in Pagoda 7D format [nk, nj, ni, kp, jp, ip, xyz]
       centres (optional, 4D numpy array of floats): cell centre points [nk, nj, ni, xyz]; calculated if None
       off_hand (boolean, default False): if True, the handedness of IJK space is the opposite of that for xyz
          space; if this argument is not set correctly, negative volumes will be returned

    returns:
       numpy 3D array of floats being the cell volumes [nk, nj, ni]

    note:
       length units are assumed to be consistent in x, y & z; and untis of returned volumes are implicitly
       those length units cubed
    """

    def pyrs(ra, ab, ac, ad, db):  # returns 6 * volume of one quad pyramid (for each cell)
        return np.sum(ra * np.cross(db, ac), axis = -1) + 0.5 * np.sum(ac * np.cross(ad, ab), axis = -1)

    # Pagoda style corner point data
    assert cp.ndim == 7

    flat = cp.reshape(-1, 2, 2, 2, 3)
    cells = flat.shape[0]
    if centres is None:
        centres = np.mean(flat, axis = (1, 2, 3))
    else:
        centres = centres.reshape((-1, 3))
    v = np.zeros(cells)
    # todo: numpy array operation covering all cells in grid

    r0 = flat[:, 0, 0, 0] - centres
    r1 = flat[:, 1, 1, 1] - centres

    v += pyrs(r0, flat[:, 0, 0, 1] - flat[:, 0, 0, 0], flat[:, 0, 1, 1] - flat[:, 0, 0, 0],
              flat[:, 0, 1, 0] - flat[:, 0, 0, 0], flat[:, 0, 1, 0] - flat[:, 0, 0, 1])
    v += pyrs(r0, flat[:, 1, 0, 0] - flat[:, 0, 0, 0], flat[:, 1, 0, 1] - flat[:, 0, 0, 0],
              flat[:, 0, 0, 1] - flat[:, 0, 0, 0], flat[:, 0, 0, 1] - flat[:, 1, 0, 0])
    v += pyrs(r0, flat[:, 0, 1, 0] - flat[:, 0, 0, 0], flat[:, 1, 1, 0] - flat[:, 0, 0, 0],
              flat[:, 1, 0, 0] - flat[:, 0, 0, 0], flat[:, 1, 0, 0] - flat[:, 0, 1, 0])
    v += pyrs(r1, flat[:, 1, 0, 1] - flat[:, 1, 1, 1], flat[:, 1, 0, 0] - flat[:, 1, 1, 1],
              flat[:, 1, 1, 0] - flat[:, 1, 1, 1], flat[:, 1, 1, 0] - flat[:, 1, 0, 1])
    v += pyrs(r1, flat[:, 1, 1, 0] - flat[:, 1, 1, 1], flat[:, 0, 1, 0] - flat[:, 1, 1, 1],
              flat[:, 0, 1, 1] - flat[:, 1, 1, 1], flat[:, 0, 1, 1] - flat[:, 1, 1, 0])
    v += pyrs(r1, flat[:, 0, 1, 1] - flat[:, 1, 1, 1], flat[:, 0, 0, 1] - flat[:, 1, 1, 1],
              flat[:, 1, 0, 1] - flat[:, 1, 1, 1], flat[:, 1, 0, 1] - flat[:, 0, 1, 1])

    v /= 6.0

    if off_hand:
        v = np.negative(v)
    return v.reshape(cp.shape[0:3])



"""High level add_zone_by_layer_property() function."""

import numpy as np

import resqpy.grid as grr
import resqpy.model as rq

from resqpy.derived_model._add_one_grid_property_array import add_one_grid_property_array
from resqpy.derived_model._zone_layer_ranges_from_array import zone_layer_ranges_from_array


def add_zone_by_layer_property(epc_file,
                               grid_uuid = None,
                               zone_by_layer_vector = None,
                               zone_by_cell_property_uuid = None,
                               use_dominant_zone = False,
                               use_local_property_kind = True,
                               null_value = -1,
                               title = 'ZONE',
                               realization = None,
                               extra_metadata = {}):
    """Adds a discrete zone property (and local property kind) with indexable element of layers.

    arguments:
       epc_file (string): file name to load resqml model from and to update with the zonal property
       grid_uuid (uuid.UUID or str, optional): required unless the model has only one grid, or one named ROOT
       zone_by_layer_vector (nk integers, optional): either this or zone_by_cell_property_uuid must be given;
          a 1D numpy array, tuple or list of ints, being the zone number to which each layer belongs
       zone_by_cell_property_uuid (uuid.UUID or str, optional): either this or zone_by_layer_vector must be given;
          the uuid of a discrete property with grid as supporting representation and cells as indexable elements,
          holidng the zone to which the cell belongs
       use_dominant_zone (boolean, default False): if True and more than one zone is represented within the cells
          of a layer, then the whole layer is assigned to the zone with the biggest count of cells in the layer;
          if False, an exception is raised if more than one zone is represented by the cells of a layer; ignored
          if zone_by_cell_property_uuid is None
       use_local_property_kind (boolean, default True): if True, the new zone by layer property is given a
          local property kind titled 'zone'; if False, the property kind will be set to 'discrete'
       null_value (int, default -1): the value to use if a layer does not belong to any zone (rarely used)
       title (str, default 'ZONE'): the citation title of the new zone by layer property
       realization (int, optional): if present the new zone by layer property is marked as belonging to this
          realization
       extra_metadata (dict, optional): any items in this dictionary are added as extra metadata to the new
          property

    returns:
       numpy vector of zone numbers (by layer), uuid of newly created property
    """

    assert zone_by_layer_vector is not None or zone_by_cell_property_uuid is not None
    assert zone_by_layer_vector is None or zone_by_cell_property_uuid is None

    model = rq.Model(epc_file)

    if grid_uuid is None:
        grid = model.grid()
        grid_uuid = grid.uuid
    else:
        grid = model.grid_for_uuid_from_grid_list(grid_uuid)
        if grid is None:
            grid = grr.any_grid(model, uuid = grid_uuid, find_properties = True)
    assert grid is not None, 'failed to establish grid object'

    if zone_by_layer_vector is not None:
        assert len(
            zone_by_layer_vector) == grid.nk, 'length of zone by layer vector does not match number of layers in grid'
        zone_by_layer = np.array(zone_by_layer_vector, dtype = int)
    elif zone_by_cell_property_uuid is not None:
        pc = grid.property_collection
        zone_by_cell_array = pc.single_array_ref(uuid = zone_by_cell_property_uuid)
        assert zone_by_cell_array is not None, 'zone by cell property array not found for uuid: ' + str(
            zone_by_cell_property_uuid)
        zone_range_list = zone_layer_ranges_from_array(zone_by_cell_array, use_dominant_zone = use_dominant_zone)
        assert zone_range_list is not None and len(
            zone_range_list) > 0, 'failed to convert zone by cell to zone by layer'
        zone_by_layer = np.full((grid.nk,), null_value, dtype = int)
        for min_k0, max_k0, zone_index in zone_range_list:
            assert 0 <= min_k0 <= max_k0 < grid.nk, 'zonal layer limits out of range for grid (probable bug)'
            zone_by_layer[min_k0:max_k0 + 1] = zone_index
    else:
        raise Exception('code failure')

    assert zone_by_layer.ndim == 1 and zone_by_layer.size == grid.nk

    if use_local_property_kind:
        property_kind = 'zone'
        zone_pk = grr.establish_zone_property_kind(model)
        local_property_kind_uuid = zone_pk.uuid
        model.store_epc(only_if_modified = True)
    else:
        property_kind = 'discrete'
        local_property_kind_uuid = None

    model.h5_release()

    property_uuid = add_one_grid_property_array(epc_file,
                                                zone_by_layer,
                                                property_kind,
                                                grid_uuid = grid.uuid,
                                                source_info = 'zone defined by layer',
                                                title = title,
                                                discrete = True,
                                                null_value = null_value,
                                                indexable_element = 'layers',
                                                realization = realization,
                                                local_property_kind_uuid = local_property_kind_uuid,
                                                extra_metadata = extra_metadata)

    return zone_by_layer, property_uuid
